// /******************************************************************************
//  * File: HandTrackingSimulationSubsystem.cs
//  * Copyright (c) 2023 Qualcomm Technologies, Inc. and/or its subsidiaries. All rights reserved.
//  *
//  * Confidential and Proprietary - Qualcomm Technologies, Inc.
//  *
//  ******************************************************************************/

using System;
using System.Collections.Generic;
using QCHT.Interactions.Core.Editor;
using QCHT.Interactions.Hands;
using UnityEngine;
using UnityEngine.InputSystem;
using UnityEngine.XR.Interaction.Toolkit.Inputs.Simulation;

namespace QCHT.Interactions.Core
{ 
  internal class HandTrackingSimulationSubsystem : XRHandTrackingSubsystem {
    
    private const string ID = "Qualcomm-HandTrackingSimulationSubsystem";

    private class HandTrackingSimulationProvider : XRHandTrackingSubsystem.Provider
    {
      private static void AssignPose(ref Pose[] joints, HandPose poseAsset, Pose wrist) {
        ConvertHandPoseToOpenXRData(ref joints, poseAsset);
        var baseOrientation = Quaternion.AngleAxis(90f, Vector3.left);
        wrist.rotation *= Quaternion.AngleAxis(90f, Vector3.right);
        for (var i = 0; i < joints.Length; i++) {
          joints[i].rotation = wrist.rotation * joints[i].rotation * baseOrientation;
          joints[i].position = wrist.position + wrist.rotation * joints[i].position;
        }
      }

      private static HandPose GetPoseAsset(XrHandGesture gesture, bool isLeft) {
        return gesture switch {
          XrHandGesture.XR_HAND_PINCH => isLeft ? XRHandSimulationHandPosesSettings.Instance.leftPinchHand : XRHandSimulationHandPosesSettings.Instance.rightPinchHand,
          XrHandGesture.XR_HAND_GRAB => isLeft ? XRHandSimulationHandPosesSettings.Instance.leftGrabHand : XRHandSimulationHandPosesSettings.Instance.rightGrabHand,
          _ => isLeft ? XRHandSimulationHandPosesSettings.Instance.leftOpenHand : XRHandSimulationHandPosesSettings.Instance.rightOpenHand
        };
      }
      
      public override void GetHandData(bool isLeft, ref bool isTracked, ref int space, ref Pose rootPose, ref Pose[] joints, ref float scale, ref int gesture, ref float gestureRatio, ref float flipRatio) {
        Status = QchtStatus.QchtStatusRunning;
        space = (int) XrSpace.XR_HAND_WORLD;
        scale = HandTrackingSimulationEditorSettings.Instance.handScale;
        
        var ctrl = isLeft ? InputSystem.GetDevice<XRSimulatedController>(CommonUsages.LeftHand) : InputSystem.GetDevice<XRSimulatedController>(CommonUsages.RightHand);
        if (ctrl == null) {
          // Shows and apply hand pose
          AssignPose(ref joints, GetPoseAsset(XrHandGesture.XR_HAND_OPEN_HAND, isLeft), Pose.identity);
          isTracked = true;
          return;
        }
        
        isTracked = ctrl.isTracked.IsPressed();
        rootPose = new Pose(ctrl.devicePosition.ReadValue(), ctrl.deviceRotation.ReadValue());
        rootPose.rotation *= Quaternion.AngleAxis(90f, Vector3.left);
        
        if (ctrl.trigger.IsPressed()) gesture = (int) XrHandGesture.XR_HAND_PINCH;
        else if (ctrl.grip.IsPressed()) gesture = (int) XrHandGesture.XR_HAND_GRAB;
        else gesture = (int) XrHandGesture.XR_HAND_OPEN_HAND;
        
        AssignPose(ref joints, GetPoseAsset((XrHandGesture) gesture, isLeft), rootPose);
      }

      private static void ConvertHandPoseToOpenXRData(ref Pose[] joints, HandPose pose) {
        // Palm
        joints[(int) XrHandJoint.XR_HAND_JOINT_PALM].position = pose.Palm.Position;
        joints[(int) XrHandJoint.XR_HAND_JOINT_PALM].rotation = pose.Palm.Rotation;
        
        // Root
        joints[(int) XrHandJoint.XR_HAND_JOINT_WRIST].position = pose.Root.Position;
        joints[(int) XrHandJoint.XR_HAND_JOINT_WRIST].rotation = pose.Root.Rotation;
        
        // Thumb
        joints[(int) XrHandJoint.XR_HAND_JOINT_THUMB_METACARPAL].position = pose.Thumb.BaseData.Position;
        joints[(int) XrHandJoint.XR_HAND_JOINT_THUMB_METACARPAL].rotation = pose.Thumb.BaseData.Rotation;
        joints[(int) XrHandJoint.XR_HAND_JOINT_THUMB_PROXIMAL].position = pose.Thumb.MiddleData.Position;
        joints[(int) XrHandJoint.XR_HAND_JOINT_THUMB_PROXIMAL].rotation = pose.Thumb.MiddleData.Rotation;
        joints[(int) XrHandJoint.XR_HAND_JOINT_THUMB_DISTAL].position = pose.Thumb.TopData.Position;
        joints[(int) XrHandJoint.XR_HAND_JOINT_THUMB_DISTAL].rotation = pose.Thumb.TopData.Rotation;
        joints[(int) XrHandJoint.XR_HAND_JOINT_THUMB_TIP].position = pose.Thumb.TopData.Position;
        joints[(int) XrHandJoint.XR_HAND_JOINT_THUMB_TIP].rotation = pose.Thumb.TopData.Rotation;
        
        // Index
        joints[(int) XrHandJoint.XR_HAND_JOINT_INDEX_METACARPAL].position = pose.Index.BaseData.Position;
        joints[(int) XrHandJoint.XR_HAND_JOINT_INDEX_METACARPAL].rotation = pose.Index.BaseData.Rotation;
        joints[(int) XrHandJoint.XR_HAND_JOINT_INDEX_PROXIMAL].position = pose.Index.BaseData.Position;
        joints[(int) XrHandJoint.XR_HAND_JOINT_INDEX_PROXIMAL].rotation = pose.Index.BaseData.Rotation;
        joints[(int) XrHandJoint.XR_HAND_JOINT_INDEX_INTERMEDIATE].position = pose.Index.MiddleData.Position;
        joints[(int) XrHandJoint.XR_HAND_JOINT_INDEX_INTERMEDIATE].rotation = pose.Index.MiddleData.Rotation;
        joints[(int) XrHandJoint.XR_HAND_JOINT_INDEX_DISTAL].position = pose.Index.TopData.Position;
        joints[(int) XrHandJoint.XR_HAND_JOINT_INDEX_DISTAL].rotation = pose.Index.TopData.Rotation;
        joints[(int) XrHandJoint.XR_HAND_JOINT_INDEX_TIP].position = pose.Index.TopData.Position;
        joints[(int) XrHandJoint.XR_HAND_JOINT_INDEX_TIP].rotation = pose.Index.TopData.Rotation;
        
        // Middle
        joints[(int) XrHandJoint.XR_HAND_JOINT_MIDDLE_METACARPAL].position = pose.Middle.BaseData.Position;
        joints[(int) XrHandJoint.XR_HAND_JOINT_MIDDLE_METACARPAL].rotation = pose.Middle.BaseData.Rotation;
        joints[(int) XrHandJoint.XR_HAND_JOINT_MIDDLE_PROXIMAL].position = pose.Middle.BaseData.Position;
        joints[(int) XrHandJoint.XR_HAND_JOINT_MIDDLE_PROXIMAL].rotation = pose.Middle.BaseData.Rotation;
        joints[(int) XrHandJoint.XR_HAND_JOINT_MIDDLE_INTERMEDIATE].position = pose.Middle.MiddleData.Position;
        joints[(int) XrHandJoint.XR_HAND_JOINT_MIDDLE_INTERMEDIATE].rotation = pose.Middle.MiddleData.Rotation;
        joints[(int) XrHandJoint.XR_HAND_JOINT_MIDDLE_DISTAL].position = pose.Middle.TopData.Position;
        joints[(int) XrHandJoint.XR_HAND_JOINT_MIDDLE_DISTAL].rotation = pose.Middle.TopData.Rotation;
        joints[(int) XrHandJoint.XR_HAND_JOINT_MIDDLE_TIP].position = pose.Middle.TopData.Position;
        joints[(int) XrHandJoint.XR_HAND_JOINT_MIDDLE_TIP].rotation = pose.Middle.TopData.Rotation;
        
        // Ring
        joints[(int) XrHandJoint.XR_HAND_JOINT_RING_METACARPAL].position = pose.Ring.BaseData.Position;
        joints[(int) XrHandJoint.XR_HAND_JOINT_RING_METACARPAL].rotation = pose.Ring.BaseData.Rotation;
        joints[(int) XrHandJoint.XR_HAND_JOINT_RING_PROXIMAL].position = pose.Ring.BaseData.Position;
        joints[(int) XrHandJoint.XR_HAND_JOINT_RING_PROXIMAL].rotation = pose.Ring.BaseData.Rotation;
        joints[(int) XrHandJoint.XR_HAND_JOINT_RING_INTERMEDIATE].position = pose.Ring.MiddleData.Position;
        joints[(int) XrHandJoint.XR_HAND_JOINT_RING_INTERMEDIATE].rotation = pose.Ring.MiddleData.Rotation;
        joints[(int) XrHandJoint.XR_HAND_JOINT_RING_DISTAL].position = pose.Ring.TopData.Position;
        joints[(int) XrHandJoint.XR_HAND_JOINT_RING_DISTAL].rotation = pose.Ring.TopData.Rotation;
        joints[(int) XrHandJoint.XR_HAND_JOINT_RING_TIP].position = pose.Ring.TopData.Position;
        joints[(int) XrHandJoint.XR_HAND_JOINT_RING_TIP].rotation = pose.Ring.TopData.Rotation;
        
        // Pinky
        joints[(int) XrHandJoint.XR_HAND_JOINT_LITTLE_METACARPAL].position = pose.Pinky.BaseData.Position;
        joints[(int) XrHandJoint.XR_HAND_JOINT_LITTLE_METACARPAL].rotation = pose.Pinky.BaseData.Rotation;
        joints[(int) XrHandJoint.XR_HAND_JOINT_LITTLE_PROXIMAL].position = pose.Pinky.BaseData.Position;
        joints[(int) XrHandJoint.XR_HAND_JOINT_LITTLE_PROXIMAL].rotation = pose.Pinky.BaseData.Rotation;
        joints[(int) XrHandJoint.XR_HAND_JOINT_LITTLE_INTERMEDIATE].position = pose.Pinky.MiddleData.Position;
        joints[(int) XrHandJoint.XR_HAND_JOINT_LITTLE_INTERMEDIATE].rotation = pose.Pinky.MiddleData.Rotation;
        joints[(int) XrHandJoint.XR_HAND_JOINT_LITTLE_DISTAL].position = pose.Pinky.TopData.Position;
        joints[(int) XrHandJoint.XR_HAND_JOINT_LITTLE_DISTAL].rotation = pose.Pinky.TopData.Rotation;
        joints[(int) XrHandJoint.XR_HAND_JOINT_LITTLE_TIP].position = pose.Pinky.TopData.Position;
        joints[(int) XrHandJoint.XR_HAND_JOINT_LITTLE_TIP].rotation = pose.Pinky.TopData.Rotation;
      }
    }
    
    private static readonly List<XRHandTrackingSubsystemDescriptor> s_handTrackingSubsystemDescriptors = new List<XRHandTrackingSubsystemDescriptor>();
    
    [RuntimeInitializeOnLoadMethod(RuntimeInitializeLoadType.SubsystemRegistration)]
    public static void RegisterDescriptor() { 
      if (!Application.isEditor || !HandTrackingSimulationEditorSettings.Instance.enabled)
        return;
      
      XRHandTrackingSubsystemDescriptor.Create(new XRHandTrackingSubsystemDescriptor.Cinfo {
        id = ID,
        providerType = typeof(HandTrackingSimulationProvider),
        subsystemTypeOverride = typeof(HandTrackingSubsystem)
      });
      
      if (HandTrackingSimulationEditorSettings.Instance.dataSource == HandTrackingSimulationEditorSettings.DataSource.SimulationSubsystem) {
        var subsystem = CreateSubsystem<XRHandTrackingSubsystemDescriptor, HandTrackingSimulationSubsystem>(s_handTrackingSubsystemDescriptors, ID);
        subsystem?.Start();
      }
    }
    
    private static ISubsystem CreateSubsystem<TDescriptor, TSubsystem>(List<TDescriptor> descriptors, string id) where TDescriptor : ISubsystemDescriptor where TSubsystem : ISubsystem {
      if (descriptors == null)
        throw new ArgumentNullException(nameof(descriptors));

      SubsystemManager.GetSubsystemDescriptors<TDescriptor>(descriptors);

      if (descriptors.Count <= 0) 
        return null;
      
      foreach (var descriptor in descriptors)
        if (string.Compare(descriptor.id, id, StringComparison.OrdinalIgnoreCase) == 0)
          return descriptor.Create();
      
      return null;
    }
  }
}