// /******************************************************************************
//  * File: ProximalInteractableEditor.cs
//  * Copyright (c) 2023 Qualcomm Technologies, Inc. and/or its subsidiaries. All rights reserved.
//  *
//  * Confidential and Proprietary - Qualcomm Technologies, Inc.
//  *
//  ******************************************************************************/

using UnityEditor;
using UnityEditor.IMGUI.Controls;
using UnityEditor.UIElements;
using UnityEngine;
using UnityEngine.UIElements;

namespace QCHT.Interactions.Proximal.Editor
{
    [CustomEditor(typeof(ProximalInteractable)), CanEditMultipleObjects]
    public class ProximalInteractableEditor : UnityEditor.Editor
    {
        private const float HandleSize = 1f;

        public VisualTreeAsset m_InspectorXML;

        private ArcHandle m_ArcHandle = new ArcHandle();
        private int p1HandleId;
        private int p2HandleId;

        private SerializedProperty _interactionTypeProperty;
        private SerializedProperty _moveValueProperty;
        private SerializedProperty _moveAxisProperty;
        private SerializedProperty _moveLimitsProperty;
        private SerializedProperty _rotationValueProperty;
        private SerializedProperty _rotationAxisProperty;
        private SerializedProperty _rotationLimitsProperty;
        private SerializedProperty _invertedProperty;

        private void OnEnable()
        {
            _interactionTypeProperty = serializedObject.FindProperty("interactionType");

            _moveValueProperty = serializedObject.FindProperty("moveValue");
            _moveAxisProperty = serializedObject.FindProperty("moveFreeAxis");
            _moveLimitsProperty = serializedObject.FindProperty("moveLimits");

            _rotationValueProperty = serializedObject.FindProperty("rotationValue");
            _rotationAxisProperty = serializedObject.FindProperty("rotationFreeAxis");
            _rotationLimitsProperty = serializedObject.FindProperty("rotationLimits");

            _invertedProperty = serializedObject.FindProperty("inverted");

            p1HandleId = GUIUtility.GetControlID(FocusType.Passive);
            p2HandleId = GUIUtility.GetControlID(FocusType.Passive);
        }

#if UNITY_2021_1_OR_NEWER
        public override VisualElement CreateInspectorGUI()
        {
            var inspector = m_InspectorXML.CloneTree();
            var interactionType = inspector.Q<EnumField>("TypeEnum");
            interactionType.RegisterValueChangedCallback(_ => RefreshUI(inspector));

            var moveBox = inspector.Q<GroupBox>("MoveBox");
            var moveFreeAxis = moveBox.Q<EnumField>("Axis");
            moveFreeAxis.RegisterValueChangedCallback(_ => RefreshUI(inspector));

            var rotationBox = inspector.Q<GroupBox>("RotationBox");
            var rotationFreeAxis = rotationBox.Q<EnumField>("Axis");
            rotationFreeAxis.RegisterValueChangedCallback(_ => RefreshUI(inspector));

            RefreshUI(inspector);
            return inspector;
        }

        private void RefreshUI(VisualElement inspector)
        {
            var interactionType = (InteractionType) _interactionTypeProperty.enumValueFlag;

            var movingPart = inspector.Q<ObjectField>("MovablePart");
            movingPart.style.display =
                interactionType == InteractionType.ConstraintMovingPart ? DisplayStyle.Flex : DisplayStyle.None;

            var moveBox = inspector.Q<GroupBox>("MoveBox");
            var rotationBox = inspector.Q<GroupBox>("RotationBox");
            moveBox.style.display = rotationBox.style.display =
                interactionType == InteractionType.ConstraintMovingPart
                    ? DisplayStyle.Flex
                    : DisplayStyle.None;

            var moveFreeAxis = (FreeAxis) _moveAxisProperty.enumValueFlag;
            var moveParam = moveBox.Q<VisualElement>("Params");
            moveParam.style.display = moveFreeAxis == FreeAxis.None ? DisplayStyle.None : DisplayStyle.Flex;

            var rotationFreeAxis = (FreeAxis) _rotationAxisProperty.enumValueFlag;
            var rotationParam = rotationBox.Q<VisualElement>("Params");
            rotationParam.style.display = rotationFreeAxis == FreeAxis.None ? DisplayStyle.None : DisplayStyle.Flex;
        }

#else
#endif

        private void OnSceneGUI()
        {
            var interactionType = (InteractionType) _interactionTypeProperty.enumValueFlag;
            if (interactionType != InteractionType.ConstraintMovingPart)
                return;

            DrawMoveAxisHandle();
            DrawRotationAxisHandle();
        }

        private void DrawMoveAxisHandle()
        {
            var moveFreeAxis = (FreeAxis) _moveAxisProperty.enumValueFlag;
            if (moveFreeAxis == FreeAxis.None)
                return;

            var interactable = target as ProximalInteractable;
            if (!interactable)
                return;


            var moveLimits = _moveLimitsProperty.vector2Value;
            var moveAxis = moveFreeAxis switch
            {
                FreeAxis.X => Vector3.right,
                FreeAxis.Y => Vector3.up,
                FreeAxis.Z => Vector3.forward,
                _ => Vector3.zero
            };

            Handles.color = moveFreeAxis.GetHandleColorFromFreeAxis();

            var p1 = moveAxis * moveLimits.x;
            var p2 = moveAxis * moveLimits.y;

            var t = interactable.MovablePart.parent;
            var position = Vector3.zero;
            var rotation = Quaternion.identity;
            
            if (t != null) // Not root
            {
                var localToWorldMatrix = t.localToWorldMatrix;
                position = t.position;
                rotation = t.rotation;
                p1 = localToWorldMatrix * p1;
                p2 = localToWorldMatrix * p2;
                p1 += position;
                p2 += position;
            }

            const float pointSize = 0.005f;
            // Free move p1
            {
                EditorGUI.BeginChangeCheck();
                var newP1 = Handles.FreeMoveHandle(p1HandleId, p1, rotation, pointSize, Vector3.zero,
                    Handles.DotHandleCap);
                if (EditorGUI.EndChangeCheck())
                {
                    Undo.RecordObject(target, "Change limits");
                    var dot = Vector3.Dot(moveAxis, newP1 - position);
                    _moveLimitsProperty.vector2Value = new Vector2(dot, moveLimits.y);
                }
            }


            // Free move p2
            EditorGUI.BeginChangeCheck();
            var newP2 = Handles.FreeMoveHandle(p2HandleId, p2, rotation, pointSize, Vector3.zero,
                Handles.DotHandleCap);
            if (EditorGUI.EndChangeCheck())
            {
                Undo.RecordObject(target, "Change limits");
                var dot = Vector3.Dot(moveAxis.normalized, newP2 - position);
                _moveLimitsProperty.vector2Value = new Vector2(moveLimits.x, dot);
            }

            Handles.DrawLine(p1, p2, HandleSize);
        }

        private void DrawRotationAxisHandle()
        {
            var rotationFreeAxis = (FreeAxis) _rotationAxisProperty.enumValueFlag;
            if (rotationFreeAxis == FreeAxis.None)
                return;

            var interactable = target as ProximalInteractable;
            if (!interactable)
                return;

            var t = interactable.transform;

            var rotationLimitsProperty = _rotationLimitsProperty;
            var inverted = _invertedProperty.boolValue;
            var rotationLimits = rotationLimitsProperty.vector2Value;
            var rotationValue = _rotationValueProperty.floatValue;
            var rotAxis = rotationFreeAxis switch
            {
                FreeAxis.X => (inverted ? -t.right : t.right, t.forward),
                FreeAxis.Y => (inverted ? -t.up : t.up, t.right),
                FreeAxis.Z => (inverted ? -t.forward : t.forward, t.up),
                _ => (Vector3.zero, Vector3.zero)
            };

            var scale = t.lossyScale;
            var from = Quaternion.AngleAxis(rotationLimits.x, rotAxis.Item1) * rotAxis.Item2;

            Handles.color = rotationFreeAxis.GetHandleColorFromFreeAxis();
            var norm = rotAxis.Item1;
            var matrix = Matrix4x4.TRS(t.position, Quaternion.LookRotation(from, norm), scale);

            using (new Handles.DrawingScope(matrix))
            {
                m_ArcHandle.angle = rotationLimits.y - rotationLimits.x;
                EditorGUI.BeginChangeCheck();
                m_ArcHandle.DrawHandle();
                if (EditorGUI.EndChangeCheck())
                {
                    Undo.RecordObject(target, "Change angle");
                    var a = Mathf.Clamp(rotationLimits.x + m_ArcHandle.angle, 0f, 360f);
                    _rotationLimitsProperty.vector2Value = new Vector2(rotationLimits.x, a);
                }
            }

            var color = Color.yellow;
            color.a = 0.1f;
            Handles.color = color;
            var angle = rotationValue * (rotationLimits.y - rotationLimits.x);
            Handles.DrawSolidArc(t.position, rotAxis.Item1, from, angle, scale.x);
        }
    }
}