// /******************************************************************************
//  * File: XRHandInteractableSnapPoseEditor.cs
//  * Copyright (c) 2023 Qualcomm Technologies, Inc. and/or its subsidiaries. All rights reserved.
//  *
//  * Confidential and Proprietary - Qualcomm Technologies, Inc.
//  *
//  ******************************************************************************/

using QCHT.Interactions.Editor;
using UnityEngine;
using UnityEngine.UIElements;
using UnityEditor;
using UnityEditor.UIElements;
using UnityEditor.SceneManagement;
using QCHT.Interactions.Hands;
using Unity.XR.CoreUtils;

namespace QCHT.Interactions.Proximal.Editor
{
    [CustomEditor(typeof(XRHandInteractableSnapPose))]
    public sealed class XRHandInteractableSnapPoseEditor : UnityEditor.Editor
    {
        private sealed class FreezeOption
        {
            public bool X;
            public bool Y;
            public bool Z;
        }

        public VisualTreeAsset m_InspectorXML;
        private VisualElement _inspector;
        
        private HandGhost _handGhost;
        [SerializeField] private GameObject leftGhost;
        [SerializeField] private GameObject rightGhost;
        
        // Edition
        private const float HandleSize = 0.005f;
        // private const float SnapAxisSize = 5f;
        private static readonly Color s_saveColor = new Color(0.321f, 0.498f, 0.611f);
        private static readonly Color s_editColor = new Color(0.321f, 0.611f, 0.325f);

        private HandData _editingData;
        private bool _editing;
        private XrHandedness _currentType;
        
        #region Editor

        private void OnDisable()
        {
            if (_handGhost)
                DestroyImmediate(_handGhost.gameObject);
        }
        
        public override VisualElement CreateInspectorGUI()
        {
            _inspector = new VisualElement();
            m_InspectorXML.CloneTree(_inspector);

            // Hand pose field
            var handPoseProperty = serializedObject.FindProperty("handPose");
            var handPoseField = _inspector.Q<ObjectField>("handPose");
            handPoseField.BindProperty(handPoseProperty);
            handPoseField.RegisterValueChangedCallback(_ =>
            {
                UpdateButtonsState();
            });
            
            UpdateButtonsState();
            
            // Edition
            const string cStop = "Save";
            const string cEdit = "Edit joints";

            var editButton = _inspector.Q<Button>("editButton");
            editButton.clicked += () =>
            {
                _editing = !_editing;
                editButton.text = _editing ? cStop : cEdit;
                editButton.style.backgroundColor = _editing ? s_editColor : s_saveColor;
                SceneView.RepaintAll();
            };

            var applyButton = _inspector.Q<Button>("applyButton");
            applyButton.clicked += () =>
            {
                var handPose = handPoseProperty.objectReferenceValue as HandPose;
                if (handPose == null) return;
                var data = handPose.ToHandData();
                handPoseProperty.objectReferenceValue = null;
                serializedObject.ApplyModifiedProperties();
                var snapPose = target as XRHandInteractableSnapPose;
                if (!snapPose) return;
                snapPose.Data = data;
                SceneView.RepaintAll();
            };

            var flipButton = _inspector.Q<Button>("flipButton");
            flipButton.clicked += () =>
            {
                var snapPose = target as XRHandInteractableSnapPose;
                if (!snapPose) return;
                var handData = snapPose.Data;
                handData.Flip();
                snapPose.Data = handData;
                SceneView.RepaintAll();
            };
        
            var exportButton = _inspector.Q<Button>("exportButton");
            exportButton.clicked += () =>
            {
                var snapPose = target as XRHandInteractableSnapPose;
                if (!snapPose) return;
                
                var handPose = CreateInstance<HandPose>();
                handPose.FromHandData(snapPose.Data);
                AssetUtils.CreateAssetInSettingsFromObj(handPose, "HandPoses", "NewHandPose");
                snapPose.HandPoseRef = handPose;
            };

            return _inspector;
        }

        private void UpdateButtonsState()
        {
            var handPoseProperty = serializedObject.FindProperty("handPose");
            var applyButton = _inspector.Q<Button>("applyButton");
            var exportButton = _inspector.Q<Button>("exportButton");

            if (handPoseProperty.objectReferenceValue as HandPose)
            {
                exportButton.style.display = DisplayStyle.None;
                applyButton.style.display = DisplayStyle.Flex;
            }
            else
            {
                exportButton.style.display = DisplayStyle.Flex;
                applyButton.style.display = DisplayStyle.None;
            }
        }
        
        private void OnSceneGUI()
        {
            var snapPose = target as XRHandInteractableSnapPose;
            if (!snapPose)
            {
                if (_handGhost)
                    DestroyImmediate(_handGhost.gameObject);
                return;
            }
            
            var snapPoseTransform = snapPose.transform;
            var data = GetSnapData();
            
            // Delete wrong ghost
            if (_handGhost && data.Handedness != _currentType)
                DestroyImmediate(_handGhost.gameObject);

            var ghost = data.Handedness == XrHandedness.XR_HAND_LEFT ? leftGhost : rightGhost;
            
            // Ghost
            if (!_handGhost && ghost)
            {
                _handGhost = InstantiateGhost(ghost, snapPose.transform);
                _currentType = data.Handedness;
                var center = snapPoseTransform.position;
                var size = Vector3.one * 0.2f;
                var bounds = new Bounds(center, size);
                SceneView.lastActiveSceneView.Frame(bounds, false);
            }
            else if (_handGhost && !ghost)
                DestroyImmediate(_handGhost.gameObject);

            if (!_handGhost)
                return;

            // Update Ghost
            _handGhost.HandPose = data;
            ApplyGhostRootPose(_handGhost, snapPoseTransform);

            if (!_editing) 
                return;
            
            UpdateJointHandles(_handGhost, snapPoseTransform.localScale.x);
            ApplyGhostToPose(_handGhost, ref data);
            ApplySnapData(data);
        }

        #endregion

        private static HandGhost InstantiateGhost(GameObject handGhost, Transform parent)
        {
            var ghost = Instantiate(handGhost).GetComponent<HandGhost>();
            var gO = ghost.gameObject;
            gO.transform.SetParent(parent);
            gO.hideFlags = HideFlags.HideAndDontSave;
            StageUtility.PlaceGameObjectInCurrentStage(gO);
            return ghost;
        }

        private static void ApplyGhostRootPose(HandGhost handGhost, Transform snapPoseTransform)
        {
            var ghostTransform = handGhost.transform;
            ghostTransform.position = snapPoseTransform.position;
            ghostTransform.rotation = snapPoseTransform.rotation;
            ghostTransform.localScale = snapPoseTransform.localScale.Divide(snapPoseTransform.lossyScale);
        }

        private static void ApplyGhostToPose(HandGhost handGhost, ref HandData data)
        {
            // Thumb
            data.Thumb.BaseData.Position = handGhost.thumbBase.localPosition;
            data.Thumb.BaseData.Rotation = handGhost.thumbBase.localRotation;
            data.Thumb.MiddleData.Position = handGhost.thumbMiddle.localPosition;
            data.Thumb.MiddleData.Rotation = handGhost.thumbMiddle.localRotation;
            data.Thumb.TopData.Position = handGhost.thumbTop.localPosition;
            data.Thumb.TopData.Rotation = handGhost.thumbTop.localRotation;

            // Index
            data.Index.BaseData.Position = handGhost.indexBase.localPosition;
            data.Index.BaseData.Rotation = handGhost.indexBase.localRotation;
            data.Index.MiddleData.Position = handGhost.indexMiddle.localPosition;
            data.Index.MiddleData.Rotation = handGhost.indexMiddle.localRotation;
            data.Index.TopData.Position = handGhost.indexTop.localPosition;
            data.Index.TopData.Rotation = handGhost.indexTop.localRotation;

            // Middle
            data.Middle.BaseData.Position = handGhost.middleBase.localPosition;
            data.Middle.BaseData.Rotation = handGhost.middleBase.localRotation;
            data.Middle.MiddleData.Position = handGhost.middleMiddle.localPosition;
            data.Middle.MiddleData.Rotation = handGhost.middleMiddle.localRotation;
            data.Middle.TopData.Position = handGhost.middleTop.localPosition;
            data.Middle.TopData.Rotation = handGhost.middleTop.localRotation;

            // Ring
            data.Ring.BaseData.Position = handGhost.ringBase.localPosition;
            data.Ring.BaseData.Rotation = handGhost.ringBase.localRotation;
            data.Ring.MiddleData.Position = handGhost.ringMiddle.localPosition;
            data.Ring.MiddleData.Rotation = handGhost.ringMiddle.localRotation;
            data.Ring.TopData.Position = handGhost.ringTop.localPosition;
            data.Ring.TopData.Rotation = handGhost.ringTop.localRotation;

            // Pinky
            data.Pinky.BaseData.Position = handGhost.pinkyBase.localPosition;
            data.Pinky.BaseData.Rotation = handGhost.pinkyBase.localRotation;
            data.Pinky.MiddleData.Position = handGhost.pinkyMiddle.localPosition;
            data.Pinky.MiddleData.Rotation = handGhost.pinkyMiddle.localRotation;
            data.Pinky.TopData.Position = handGhost.pinkyTop.localPosition;
            data.Pinky.TopData.Rotation = handGhost.pinkyTop.localRotation;
        }

        private void ApplySnapData(HandData data)
        {
            var snapPose = target as XRHandInteractableSnapPose;
            if (!snapPose) return;
            snapPose.Data = data;
            EditorUtility.SetDirty(snapPose);
        }

        private HandData GetSnapData()
        {
            var snapPose = target as XRHandInteractableSnapPose;
            if (!snapPose) return HandData.Default;
            var snapDataProperty = serializedObject.FindProperty("handPose");
            var handPose = snapDataProperty.objectReferenceValue as HandPose;
            return handPose ? handPose.ToHandData() : snapPose.Data;    
        }
        
        private static void UpdateJointHandles(HandGhost handGhost, float scale)
        {
            // Thumb 
            UpdateJointHandle(handGhost.thumbBase, scale, new FreezeOption ());
            UpdateJointHandle(handGhost.thumbMiddle, scale, new FreezeOption {Y = true, Z = true});
            UpdateJointHandle(handGhost.thumbTop, scale, new FreezeOption {Y = true, Z = true});

            // Index 
            UpdateJointHandle(handGhost.indexBase, scale, new FreezeOption {Z = true});
            UpdateJointHandle(handGhost.indexMiddle, scale, new FreezeOption {Y = true, Z = true});
            UpdateJointHandle(handGhost.indexTop, scale, new FreezeOption {Y = true, Z = true});

            // Middle
            UpdateJointHandle(handGhost.middleBase, scale, new FreezeOption {Z = true});
            UpdateJointHandle(handGhost.middleMiddle, scale, new FreezeOption {Y = true, Z = true});
            UpdateJointHandle(handGhost.middleTop, scale, new FreezeOption {Y = true, Z = true});

            // Ring
            UpdateJointHandle(handGhost.ringBase, scale, new FreezeOption {Z = true});
            UpdateJointHandle(handGhost.ringMiddle, scale, new FreezeOption {Y = true, Z = true});
            UpdateJointHandle(handGhost.ringTop, scale, new FreezeOption {Y = true, Z = true});

            // Pinky
            UpdateJointHandle(handGhost.pinkyBase, scale, new FreezeOption {Z = true});
            UpdateJointHandle(handGhost.pinkyMiddle, scale, new FreezeOption {Y = true, Z = true});
            UpdateJointHandle(handGhost.pinkyTop, scale, new FreezeOption {Y = true, Z = true});
        }

        private static void UpdateJointHandle(Transform transform, float scale, FreezeOption freeze = null)
        {
            if (freeze is not {X: true})
                UpdateAxisDiscHandle(transform, transform.right, Handles.xAxisColor, scale);

            if (freeze is not {Y: true})
                UpdateAxisDiscHandle(transform, transform.forward, Handles.yAxisColor, scale);

            if (freeze is not {Z: true})
                UpdateAxisDiscHandle(transform, transform.up, Handles.zAxisColor, scale);
        }

        private static void UpdateAxisDiscHandle(Transform transform, Vector3 dir, Color color, float scale)
        {
            Handles.color = color;
            transform.rotation = Handles.Disc(transform.rotation, transform.position, dir, HandleSize * scale, false, 0f);
        }
    }
}