// /******************************************************************************
//  * File: XRHandInteractableSnapPoseProviderEditor.cs
//  * Copyright (c) 2023 Qualcomm Technologies, Inc. and/or its subsidiaries. All rights reserved.
//  *
//  * Confidential and Proprietary - Qualcomm Technologies, Inc.
//  *
//  ******************************************************************************/

using QCHT.Interactions.Extensions;
using UnityEngine;
using UnityEngine.UIElements;
using UnityEditor;
using UnityEditor.SceneManagement;
using QCHT.Interactions.Hands;

namespace QCHT.Interactions.Proximal.Editor
{
    [CustomEditor(typeof(XRHandInteractableSnapPoseProvider))]
    public sealed class XRHandInteractableSnapPoseProviderEditor : UnityEditor.Editor
    {
        private VisualElement _inspector;

        [SerializeField] private GameObject leftGhost;
        [SerializeField] private GameObject rightGhost;
        
        private HandGhost _handGhost;
        
        private XrHandedness _currentType;
        
        private HandData _interpolatedHandPose;
        private Matrix4x4 _interpolatedRootPose;

        private float _scale = 1f;
        
        #region Editor
        private void OnDisable()
        {
            if (_handGhost)
                DestroyImmediate(_handGhost.gameObject);
        }

        public override void OnInspectorGUI()
        {
            base.OnInspectorGUI();
            EditorGUI.BeginChangeCheck();
            var posesProperty = serializedObject.FindProperty("poses");
            if (posesProperty.arraySize <= 0)
                return;

            XRHandInteractableSnapPose minPose = null;
            XRHandInteractableSnapPose maxPose = null;
            
            for (var i = 0; i < posesProperty.arraySize; i++)
            {
                var pose = posesProperty.GetArrayElementAtIndex(i).objectReferenceValue as XRHandInteractableSnapPose;
                if (pose == null) continue;
                if (minPose == null) minPose = pose;
                if (maxPose == null) maxPose = pose;
                if (minPose.Scale > pose.Scale) minPose = pose;
                if (maxPose.Scale < pose.Scale) maxPose = pose;
            }
            
            if (minPose != null && maxPose != null && Mathf.Abs(minPose.Scale - maxPose.Scale) > Mathf.Epsilon)
                _scale = EditorGUILayout.Slider("Scale", _scale, minPose.Scale, maxPose.Scale);
            
            if (EditorGUI.EndChangeCheck())
                SceneView.RepaintAll();
        }

        private void OnSceneGUI()
        {
            var provider = target as XRHandInteractableSnapPoseProvider;
            
            if (!provider)
            {
                if (_handGhost) DestroyImmediate(_handGhost.gameObject);
                return;
            }
            
            if (!provider.TryGetInterpolatedHandPoseFromScale(ref _interpolatedHandPose, ref _interpolatedRootPose, _scale))
                return;
            
            var providerTransform = provider.transform;

            // Delete wrong ghost 
            if (_handGhost && _handGhost.HandPose.Handedness != _currentType)
                DestroyImmediate(_handGhost.gameObject);

            var ghost = _interpolatedHandPose.Handedness == XrHandedness.XR_HAND_LEFT ? leftGhost : rightGhost;
            
            // Ghost
            if (!_handGhost && ghost)
            {
                _handGhost = InstantiateGhost(ghost, providerTransform);
                _currentType = _interpolatedHandPose.Handedness;
                var center = providerTransform.position;
                var size = Vector3.one * 0.2f;
                var bounds = new Bounds(center, size);
                SceneView.lastActiveSceneView.Frame(bounds, false);
            }
            else if (_handGhost && !ghost)
                DestroyImmediate(_handGhost.gameObject);

            if (!_handGhost)
                return;

            // Update Ghost
            _handGhost.HandPose = _interpolatedHandPose;
            ApplyGhostRootPose(_handGhost, _interpolatedRootPose, providerTransform);
        }

        #endregion

        private static HandGhost InstantiateGhost(GameObject handObj, Transform parent)
        {
            var ghost = Instantiate(handObj).GetComponent<HandGhost>();
            var gO = ghost.gameObject;
            gO.transform.SetParent(parent);
            gO.hideFlags = HideFlags.HideAndDontSave;
            StageUtility.PlaceGameObjectInCurrentStage(gO);
            return ghost;
        }

        private static void ApplyGhostRootPose(HandGhost handGhost, Matrix4x4 rootPose, Transform providerTransform)
        {
            var ghostTransform = handGhost.transform;
            ghostTransform.localRotation = rootPose.rotation;
            ghostTransform.localPosition = rootPose.GetPosition();
            ghostTransform.localScale = rootPose.lossyScale.Divide(providerTransform.lossyScale);
        }
    }
}