// /******************************************************************************
//  * File: XRHandTrackingManager.cs
//  * Copyright (c) 2023 Qualcomm Technologies, Inc. and/or its subsidiaries. All rights reserved.
//  *
//  * Confidential and Proprietary - Qualcomm Technologies, Inc.
//  *
//  ******************************************************************************/

using UnityEngine;
using UnityEngine.InputSystem;
using UnityEngine.InputSystem.LowLevel;
using UnityEngine.XR;
using QCHT.Interactions.Hands;
using QCHT.Interactions.Hands.VFF;
using UnityEngine.XR.Interaction.Toolkit;
using CommonUsages = UnityEngine.InputSystem.CommonUsages;
using InputDevice = UnityEngine.InputSystem.InputDevice;

#if UNITY_EDITOR
using UnityEditor;
#endif

namespace QCHT.Interactions.Core
{
    [DisallowMultipleComponent]
    [DefaultExecutionOrder(XRInteractionUpdateOrder.k_Controllers)]
    public sealed class XRHandTrackingManager : XRSubsystemLifeCycleManager<XRHandTrackingSubsystem,
        XRHandTrackingSubsystemDescriptor, XRHandTrackingSubsystem.Provider>
    {
        private const string kResourcesHandLeft = "QualcommHandLeft";
        private const string kResourcesHandRight = "QualcommHandRight";
        private const string kTrackableLeftName = "QC Hand Left";
        private const string kTrackableRightName = "QC Hand Right";
        
        /// <summary>
        /// Represents phases during data could be retrieved.
        /// </summary>
        public enum UpdatePhase {
            UpdateAndBeforeRender,
            Update,
            BeforeRender,
            Fixed
        }
        
        /// <summary>
        /// When the hand tracking data should be updated?
        /// </summary>
        [SerializeField] private UpdatePhase updatePhase = UpdatePhase.UpdateAndBeforeRender;

        [Space]
        
        [SerializeField] private GameObject leftHandPrefab;
        [SerializeField] private GameObject rightHandPrefab;

        [Space]

        private static readonly Vector3 s_positionOffsetLeft = new Vector3(0.02f, -0.09f, 0.1f);
        private static readonly Vector3 s_positionOffsetRight = new Vector3(-0.02f, -0.09f, 0.1f);

        public UpdatePhase UpdateTrackingPhase {
            get => updatePhase;
            set => updatePhase = value;
        }
        
        public XRHandTrackingSubsystem.QchtStatus HandTrackingStatus => subsystem?.Status ?? XRHandTrackingSubsystem.QchtStatus.QchtStatusIdle;
        
        private GameObject _leftHand;
        private GameObject _rightHand;

        private HandTrackingDevice _leftHandDevice;
        private HandTrackingDevice _rightHandDevice;

        private HandTrackingInputState _leftState;
        private HandTrackingInputState _rightState;

        private Pose[] _leftPoses = new Pose[(int)XrHandJoint.XR_HAND_JOINT_MAX];
        private Pose[] _rightPoses = new Pose[(int)XrHandJoint.XR_HAND_JOINT_MAX];

        private Transform _origin;
        private Camera _camera;
        
        /// <summary>
        /// Returns the current left hand game object instance.
        /// </summary>
        public GameObject LeftHand => _leftHand;
        
        /// <summary>
        /// Returns the current right hand game object instance.
        /// </summary>
        public GameObject RightHand => _rightHand;
        
        /// <summary>
        /// Left prefab object that will be instantiated.
        /// If the prefab changed after Instantiation time, it can be refreshed by calling RefreshLeftHand
        /// </summary>
        public GameObject LeftHandPrefab {
            get => leftHandPrefab;
            set => leftHandPrefab = value;
        }
        
        /// <summary>
        /// Right prefab object that will be instantiated.
        /// If the prefab changed after Instantiation time, it can be refreshed by calling RefreshRightHand.
        /// </summary>
        public GameObject RightHandPrefab {
            get => rightHandPrefab;
            set => rightHandPrefab = value;
        }
        
        private void Awake() {
            _origin = XROriginUtility.GetOriginTransform();
            _camera = XROriginUtility.GetOriginCamera();
        }
        
        protected override void OnEnable() {
            base.OnEnable();
            StartDevices();
            Application.onBeforeRender += OnBeforeRender;
        }

        protected override void OnDisable() {
            base.OnDisable();
            StopDevices();
            Application.onBeforeRender -= OnBeforeRender;
        }

        protected override void OnDestroy() {
            base.OnDestroy();
            StopDevices();
            if (_leftHand != null) DestroyImmediate(_leftHand);
            if (_rightHand != null) DestroyImmediate(_rightHand);
        }

        private void Start() {
            InstantiateRightHand();
            InstantiateLeftHand();
            ToggleLeftHand(_leftState.isTracked);
            ToggleRightHand(_rightState.isTracked);
        }
        
        [BeforeRenderOrder(XRInteractionUpdateOrder.k_Controllers)]
        private void OnBeforeRender() {
            if (updatePhase == UpdatePhase.BeforeRender ||
                updatePhase == UpdatePhase.UpdateAndBeforeRender) {
                UpdateHands();      
            }
        }

        private void FixedUpdate() {
            if (updatePhase == UpdatePhase.Fixed) {
                UpdateHands();
            }
        }

        private void Update() {
            if (updatePhase == UpdatePhase.Update ||
                updatePhase == UpdatePhase.UpdateAndBeforeRender) {
                UpdateHands();
            }
        }

        #region Input devices
        
        /// <summary>
        /// Adds hand tracking devices in the input devices list.
        /// Devices are in the input devices list during all the lifetime of the Hand Tracking Manager.
        /// See StartDevices and StopDevices to 
        /// </summary>
        private void AddDevices() {
            _leftHandDevice ??= AddHandDevice(true) as HandTrackingDevice;
            _rightHandDevice ??= AddHandDevice(false) as HandTrackingDevice;
        }

        /// <summary>
        /// Removes hand tracking devices from the input system.
        /// </summary>
        private void RemoveDevices() {
            RemoveHandDevice(_leftHandDevice);
            _leftHandDevice = null;
            RemoveHandDevice(_rightHandDevice);
            _rightHandDevice = null;
        }
        
        /// <summary>
        /// Enables hand tracking devices in the input system.
        /// </summary>
        public void StartDevices() {
            if (subsystem == null) return; 
            AddDevices();
            ToggleLeftHand(_leftState.isTracked);
            ToggleRightHand(_rightState.isTracked);
        }
        
        /// <summary>
        /// Disables hand tracking devices in the input system.
        /// </summary>
        public void StopDevices() {
            ToggleLeftHand(false);
            ToggleRightHand(false);
            RemoveDevices();
        }
        
        /// <summary>
        /// Updates hand tracking devices.
        /// </summary>
        private void UpdateHands() {
            UpdateHand(ref _leftState, true, _leftHand, _leftHandDevice, ref _leftPoses);
            UpdateHand(ref _rightState, false, _rightHand, _rightHandDevice, ref _rightPoses);
        } 
        
        #endregion
        
        /// <summary>
        /// Update the hand device and objects,
        /// It will update the InputState struct for QCHT hand tracked device.
        /// That should be done by OpenXR layer and retrieved as a controller profile in future versions
        /// </summary>
        private void UpdateHand(ref HandTrackingInputState state, bool isLeft, GameObject hand, HandTrackingDevice device, ref Pose[] joints) {
            var isTracked = false;  
            var gestureId = 0;
            var gestureRatio = 0f;
            var flipRatio = -1f;
            var space = (int) XrSpace.XR_HAND_WORLD;
            var rootPose = new Pose();
            var scale = 1f;
            
            // Get hand tracking data
            if (subsystem != null && subsystem.running)
                subsystem.GetHandData(isLeft, ref isTracked, ref space, ref rootPose, ref joints, ref scale, ref gestureId, ref gestureRatio, ref flipRatio);
            
            if (isTracked != state.isTracked) {
                if (isLeft) ToggleLeftHand(isTracked);
                else ToggleRightHand(isTracked);
            }
            
            if ((XrSpace) space == XrSpace.XR_HAND_WORLD)
                ConvertInOrigin(ref joints);
            
            // Update hand tracked objects
            if (TryGetComponentOnHand(hand, out IHandDataUpdater dataUpdater)) {
                dataUpdater.UpdateData((XrSpace) space, joints, scale);
            }
            
            if (device == null) return;
            
            state.isTracked = isTracked;
            if (state.isTracked) {
                state.trackingState = (int) (InputTrackingState.Position | InputTrackingState.Rotation);
                var camT = _camera.transform;
                var cameraPose = new Pose(camT.localPosition, camT.localRotation);
                var devicePose = Unity.XR.CoreUtils.TransformExtensions.InverseTransformPose(_origin, rootPose);
                devicePose.position = _origin.TransformPoint(devicePose.position); 
                devicePose.rotation = _origin.rotation * devicePose.rotation;

                // Device pose
                state.devicePosition = devicePose.position;
                state.deviceRotation = devicePose.rotation;
                
                // Pointer pose
                const float pointerSmoothValue = 7.5f;
                const float shoulderOffset = 0.08f;
                var positionOffset = isLeft ? s_positionOffsetLeft : s_positionOffsetRight;
                positionOffset *= scale;
                var posOffset = devicePose.rotation * positionOffset;
                var shoulder = isLeft ? -cameraPose.right : cameraPose.right;
                var shoulderPosition = cameraPose.position + (-cameraPose.up + shoulder) * shoulderOffset;
                var pointerPose = Pose.identity;
                pointerPose.position = devicePose.position + posOffset;
                pointerPose.rotation = Quaternion.LookRotation((pointerPose.position - shoulderPosition).normalized);
                
                // Poke position
                var indexPose = joints[(int) XrHandJoint.XR_HAND_JOINT_INDEX_TIP];
                state.pokePosition = Quaternion.Inverse(devicePose.rotation) * (indexPose.position - devicePose.position);
                state.pokeRotation = Quaternion.Inverse(devicePose.rotation) * indexPose.rotation;
                state.pokePosition *= scale;
                
                var dt = Time.deltaTime * pointerSmoothValue;
                state.pointerPosition = Vector3.Lerp(state.pointerPosition, pointerPose.position, dt);
                state.pointerRotation = Quaternion.Slerp(state.pointerRotation, pointerPose.rotation, dt);

                // Trigger pressed
                // var releaseThreshold = handRiggedVisualizer.HandPose.GetMinDistToThumbTip();
                // state.selectPressed = releaseThreshold < 0.055f;
                state.triggerPressed = (XrHandGesture) gestureId == XrHandGesture.XR_HAND_PINCH;
                state.trigger = (XrHandGesture) gestureId == XrHandGesture.XR_HAND_PINCH ? 1f : 0f;
                state.gripPressed = (XrHandGesture) gestureId == XrHandGesture.XR_HAND_GRAB;
                state.grip = (XrHandGesture) gestureId == XrHandGesture.XR_HAND_GRAB ? 1f : 0f;

                state.flipRatio = flipRatio;
            }
            
            InputState.Change(device, state);
        }

        /// <summary>
        /// Converts joints in xr origin space
        /// </summary>
        /// <param name="joints"> joints to convert </param>
        private void ConvertInOrigin(ref Pose[] joints) {
            for (var i = 0; i < joints.Length; i++) {
                var openXRPose = joints[i];
                openXRPose.position = _origin.TransformPoint(openXRPose.position);
                openXRPose.rotation = _origin.rotation * openXRPose.rotation;
                joints[i] = openXRPose;
            }
        }
        
        #region Hand Objects

        /// <summary>
        /// Forces regenerating the left hand trackable object.
        /// </summary>
        public void RefreshLeftHand() {
            RemoveLeftHand();
            InstantiateLeftHand();
            ToggleLeftHand(_leftState.isTracked);
        }
        
        /// <summary>
        /// Forces regenerating the right hand trackable object.
        /// </summary>
        public void RefreshRightHand() {
            RemoveRightHand();
            InstantiateRightHand();
            ToggleRightHand(_rightState.isTracked);
        }
        
        /// <summary>
        /// Instantiates left hand trackable object.
        /// </summary>
        private void InstantiateLeftHand() => InstantiateHand(kTrackableLeftName, leftHandPrefab, ref _leftHand);
        
        /// <summary>
        /// Instantiates right hand trackable object.
        /// </summary>
        private void InstantiateRightHand() => InstantiateHand(kTrackableRightName, rightHandPrefab, ref _rightHand);

        /// <summary>
        /// Instantiate a hand object.
        /// </summary>
        private void InstantiateHand(string objectName, GameObject prefab, ref GameObject hand) {
            if (hand == null) {
                hand = InstantiateHandTrackable(objectName, prefab);
            }

            if (hand != null) {
                SetParentTrackable(hand);
            }
        }

        /// <summary>
        /// Destroys left hand trackable object.
        /// </summary>
        private void RemoveLeftHand() => RemoveHand(ref _leftHand);

        /// <summary>
        /// Destroys right hand trackable object.
        /// </summary>
        private void RemoveRightHand() => RemoveHand(ref _rightHand);

        /// <summary>
        /// Destroys a hand object.
        /// </summary>
        private static void RemoveHand(ref GameObject hand) {
            if (hand != null) Destroy(hand);
            hand = null;
        }

        /// <summary>
        /// Sets the left hand skin.
        /// Do nothing if the hand object is not ISkinnable.
        /// </summary>
        /// <param name="skin"> Skin to set. </param>
        public void SetLeftHandSkin(HandSkin skin) => TrySetSkin(_leftHand, skin);
        
        /// <summary>
        /// Sets the right hand skin.
        /// Do nothing if the hand object is not ISkinnable.
        /// </summary>
        /// <param name="skin"> Skin to set. </param>
        public void SetRightHandSkin(HandSkin skin) => TrySetSkin(_rightHand, skin);

        private static bool TrySetSkin(GameObject hand, HandSkin skin) {
            if (TryGetComponentOnHand(hand, out IHandSkinnable skinnable)) {
                skinnable.SetSkin(skin);
                return true;
            }
            Debug.Log("[XRHandTrackingManager:TrySetSkin] Hand object is not skinnable.");
            return false;
        }
        
        /// <summary>
        /// Toggles the left hand visibility.
        /// Performs fading if the hand object is a IHandFadable. 
        /// </summary>
        /// <param name="visible"> Is the object visible? </param>
        public void ToggleLeftHand(bool visible) => ToggleHand(_leftHand, visible);

        /// <summary>
        /// Toggles the right hand visibility.
        /// </summary>
        /// <param name="visible"> Is the object visible? </param>
        public void ToggleRightHand(bool visible) => ToggleHand(_rightHand, visible);

        /// <summary>
        /// Toggles a hand game object visibility.
        /// </summary>
        private void ToggleHand(GameObject hand, bool visible) { 
            if (hand == null) return;
            if (visible) {
                hand.gameObject.SetActive(true);
                TryShow(hand);
            }
            else {
                TryHide(hand);
                hand.gameObject.SetActive(false);
            }
        }
        
        private static bool TryShow(GameObject hand) {
            if (TryGetComponentOnHand(hand, out IHideable hideable)) {
                hideable.Show();
                return true;
            }
            return false;
        }

        private static bool TryHide(GameObject hand) {
            if (TryGetComponentOnHand(hand, out IHideable hideable)) {
                hideable.Hide();
                return true;
            }
            return false;
        }
        
        /// <summary>
        /// Sets virtual force feed back system state if available on left hand object.
        /// </summary>
        /// <param name="active"> is vff active? </param>
        public void SetVffLeft(bool active) => TrySetVff(_leftHand, active);

        /// <summary>
        /// Sets virtual force feed back system state if available on right hand object.
        /// </summary>
        /// <param name="active"> is vff active? </param>
        public void SetVffRight(bool active) => TrySetVff(_rightHand, active);
        
        private static bool TrySetVff(GameObject hand, bool active) {
            if (TryGetComponentOnHand(hand, out IPhysible vffable)) {
                vffable.SetPhysible(active);
                return true;
            }
            return false;
        }
        
        private static bool TryGetComponentOnHand<T>(GameObject hand, out T component) {
            if (hand == null) {
                component = default;
                return false;
            }

            // Check component on XRBaseController model
            if (hand.TryGetComponent(out XRBaseController xrController) && xrController.model != null &&
                xrController.model.TryGetComponent(out component))
                return true;
            
            return hand.TryGetComponent(out component);
        }
        
        private void SetParentTrackable(GameObject hand) {
            if (hand == null) return;
            var trackablesParent = XROriginUtility.GetTrackableParent();
            trackablesParent = trackablesParent ? trackablesParent : transform;
            var handTransform = hand.transform;
            handTransform.SetParent(trackablesParent);
            handTransform.localPosition = Vector3.zero;
            handTransform.localRotation = Quaternion.identity;
        }
        
        #endregion

        #region static

        public static GameObject DefaultLeftHandPrefab => Resources.Load<GameObject>(kResourcesHandLeft);
        public static GameObject DefaultRightHandPrefab => Resources.Load<GameObject>(kResourcesHandRight);

        /// <summary>
        /// Instantiates a hand tracking manager with default hands prefabs.
        /// </summary>
        public static XRHandTrackingManager InstantiateHandTrackingManager() {
            return GetOrCreate(DefaultLeftHandPrefab, DefaultRightHandPrefab);
        }
        
        /// <summary>
        /// Gets existing or creates a Hand Tracking Manager if it doesn't exist.
        /// </summary>
        /// <returns> New or existing hand tracking manager instance. </returns>
        public static XRHandTrackingManager GetOrCreate(GameObject leftPrefab = null, GameObject rightPrefab = null) {
            var origin = XROriginUtility.GetOriginTransform();
            if (origin != null) {
                var manager = origin.gameObject.AddComponent<XRHandTrackingManager>();
                var needRefreshL = true;
                var needRefreshR = true;
                if (manager == null) { // if already exists
                    manager = origin.gameObject.GetComponent<XRHandTrackingManager>();
                    needRefreshL = leftPrefab && leftPrefab != manager.LeftHandPrefab;
                    needRefreshR = rightPrefab && rightPrefab != manager.RightHandPrefab;
                }
                manager.leftHandPrefab = leftPrefab;
                manager.rightHandPrefab = rightPrefab;
                if (needRefreshL) {
                    manager.RefreshLeftHand();
                }
                if (needRefreshR) {
                    manager.RefreshRightHand();
                }
#if UNITY_EDITOR   
                Selection.activeGameObject = origin.gameObject;
#endif
                return manager;
            }

            Debug.LogWarning("[XRHandTrackingManager] Hand tracking manager may not work correctly because it is not related to a ARSession/XROrigin");
            var go = new GameObject("HandTrackingManager");
            return go.AddComponent<XRHandTrackingManager>();
        }
        
        /// <summary>
        /// Destroys Hand Tracking Manager instance if it does exist.
        /// </summary>
        public static void Destroy(XRHandTrackingManager manager = null) {
            manager = manager ? manager : FindObjectOfType<XRHandTrackingManager>();
            if (manager == null) return;
            GameObject.Destroy(manager);
        }
        
        private static GameObject InstantiateHandTrackable(string handName, GameObject prefab) {
            if (prefab == null) {
                return null;
            }

            GameObject instance;
            if (prefab.scene.rootCount == 0) {
                instance = Instantiate(prefab);
                instance.name = handName;
            }
            else {
                instance = prefab;
            }
            return instance;
        }
        
        private static InputDevice AddHandDevice(bool isLeft) {
            var usage = isLeft ? CommonUsages.LeftHand : CommonUsages.RightHand;
            var device = InputSystem.AddDevice<HandTrackingDevice>($"{nameof(HandTrackingDevice)} - {usage}");
            if (device != null) InputSystem.SetDeviceUsage(device, usage);
            return device;
        }

        private static void RemoveHandDevice(InputDevice device) {
            if (device == null) return;
            InputSystem.RemoveDevice(device);
        }

        #endregion 
    }
}