// /******************************************************************************
//  * File: HandDriver.cs
//  * Copyright (c) 2023 Qualcomm Technologies, Inc. and/or its subsidiaries. All rights reserved.
//  *
//  * Confidential and Proprietary - Qualcomm Technologies, Inc.
//  *
//  ******************************************************************************/

using System;
using System.Collections.Generic;
using System.Text;
using UnityEngine;
using QCHT.Interactions.Core;
using Unity.XR.CoreUtils;

namespace QCHT.Interactions.Hands
{
    public sealed class HandDriver : MonoBehaviour, IHandDataUpdater, IHandPoseOverrider
    {
        public HandData? HandPoseOverride { get; set; }
        public HandMask? HandPoseMask { get; set; }

        [Header("Joints")]
        [SerializeField] private HandJointUpdater rootPart;

        [Space]

        // Thumb
        [SerializeField] private HandJointUpdater thumbBase;
        [SerializeField] private HandJointUpdater thumbMiddle;
        [SerializeField] private HandJointUpdater thumbTop;

        [Space]

        // Index
        [SerializeField] private HandJointUpdater indexBase;
        [SerializeField] private HandJointUpdater indexMiddle;
        [SerializeField] private HandJointUpdater indexTop;

        [Space]

        // Middle
        [SerializeField] private HandJointUpdater middleBase;
        [SerializeField] private HandJointUpdater middleMiddle;
        [SerializeField] private HandJointUpdater middleTop;

        [Space]

        // Ring
        [SerializeField] private HandJointUpdater ringBase;
        [SerializeField] private HandJointUpdater ringMiddle;
        [SerializeField] private HandJointUpdater ringTop;

        [Space]

        // Pinky
        [SerializeField] private HandJointUpdater pinkyBase;
        [SerializeField] private HandJointUpdater pinkyMiddle;
        [SerializeField] private HandJointUpdater pinkyTop;

        public Transform RootTransform => rootPart.transform;
        public HandJointUpdater RootPart => rootPart;
        public Pose RootPose => rootPart.transform.GetWorldPose();
        public float RootScale {
            get => rootPart.transform.localScale.x;
            private set => rootPart.transform.localScale = value * Vector3.one;
        }

        #region IHandDataUpdater

        public void UpdateData(XrSpace space, IReadOnlyList<Pose> joints, float scale)
        {
            UpdateRoot(space, joints, scale);
            UpdateFingerJoints(space, joints, XrFinger.XR_HAND_THUMB);
            UpdateFingerJoints(space, joints, XrFinger.XR_HAND_INDEX);
            UpdateFingerJoints(space, joints, XrFinger.XR_HAND_MIDDLE);
            UpdateFingerJoints(space, joints, XrFinger.XR_HAND_RING);
            UpdateFingerJoints(space, joints, XrFinger.XR_HAND_PINKY);
        }

        #endregion

        private void UpdateRoot(XrSpace space, IReadOnlyList<Pose> joints, float scale) 
        {
            RootScale = HandPoseOverride != null ? RootScale : scale;
            var baseOrientation = Quaternion.AngleAxis(90f, Vector3.right);
            var rootBone = new BoneData {
                Rotation = joints[(int) XrHandJoint.XR_HAND_JOINT_WRIST].rotation * baseOrientation,
                Position = joints[(int) XrHandJoint.XR_HAND_JOINT_WRIST].position,
                UpdateRotation = !HandPoseOverride.HasValue,
                // UpdatePosition = !HandPoseOverride.HasValue
            };
            rootPart.UpdateJoint(space, rootBone);
        }
        
        private void UpdateFingerJoints(XrSpace space, IReadOnlyList<Pose> joints, XrFinger id)
        {
            var maskState = GetMaskStateForFinger(id);
            var (bottom, middle, top) = GetJointUpdatersForFinger(id);
            var fingerData = maskState == HandMaskState.Required ? GetFingerDataFromHandPoseOverride(id) : GetFingerDataFromJoints(joints, id);
            if (HandPoseOverride != null) space = maskState == HandMaskState.Required ? HandPoseOverride.Value.Space : space;
            bottom?.UpdateJoint(space, fingerData.BaseData);
            middle?.UpdateJoint(space, fingerData.MiddleData);
            top?.UpdateJoint(space, fingerData.TopData);
        }

        private HandMaskState GetMaskStateForFinger(XrFinger id)
        {
            if (HandPoseMask == null)
                return HandMaskState.Free;

            return id switch {
                XrFinger.XR_HAND_THUMB => HandPoseMask.Value.Thumb,
                XrFinger.XR_HAND_INDEX => HandPoseMask.Value.Index,
                XrFinger.XR_HAND_MIDDLE => HandPoseMask.Value.Middle,
                XrFinger.XR_HAND_RING => HandPoseMask.Value.Ring,
                XrFinger.XR_HAND_PINKY => HandPoseMask.Value.Pinky,
                _ => throw new ArgumentOutOfRangeException(nameof(id), id, null)
            };
        }

        private FingerData GetFingerDataFromHandPoseOverride(XrFinger id)
        {
            if (HandPoseOverride == null)
                return FingerData.Default;
            
            return id switch {
                XrFinger.XR_HAND_THUMB => HandPoseOverride.Value.Thumb,
                XrFinger.XR_HAND_INDEX => HandPoseOverride.Value.Index,
                XrFinger.XR_HAND_MIDDLE => HandPoseOverride.Value.Middle,
                XrFinger.XR_HAND_RING => HandPoseOverride.Value.Ring,
                XrFinger.XR_HAND_PINKY => HandPoseOverride.Value.Pinky,
                _ => throw new ArgumentOutOfRangeException(nameof(id), id, null)
            };
        }

        private FingerData GetFingerDataFromJoints(IReadOnlyList<Pose> joints, XrFinger finger)
        {
            var baseOrientation = Quaternion.AngleAxis(90f, Vector3.right);
            var fingerData = finger == XrFinger.XR_HAND_THUMB ? FingerData.DefaultThumb : FingerData.Default;
            var (baseJoint, middleJoint, topJoint) = GetJointsForFinger(finger);
            fingerData.BaseData.Position = joints[(int) baseJoint].position;
            fingerData.BaseData.Rotation = joints[(int) baseJoint].rotation * baseOrientation;
            fingerData.MiddleData.Position = joints[(int) middleJoint].position;
            fingerData.MiddleData.Rotation = joints[(int) middleJoint].rotation * baseOrientation;
            fingerData.TopData.Position = joints[(int) topJoint].position;
            fingerData.TopData.Rotation = joints[(int) topJoint].rotation * baseOrientation;
            return fingerData;
        }

        private (IHandJointUpdater, IHandJointUpdater, IHandJointUpdater) GetJointUpdatersForFinger(XrFinger id) =>
            id switch {
                XrFinger.XR_HAND_THUMB => (thumbBase, thumbMiddle, thumbTop),
                XrFinger.XR_HAND_INDEX => (indexBase, indexMiddle, indexTop),
                XrFinger.XR_HAND_MIDDLE => (middleBase, middleMiddle, middleTop),
                XrFinger.XR_HAND_RING => (ringBase, ringMiddle, ringTop),
                XrFinger.XR_HAND_PINKY => (pinkyBase, pinkyMiddle, pinkyTop),
                _ => throw new ArgumentOutOfRangeException(nameof(id), id, null)
            };

        private static (XrHandJoint, XrHandJoint, XrHandJoint) GetJointsForFinger(XrFinger id) =>
            id switch
            {
                XrFinger.XR_HAND_THUMB => (
                    XrHandJoint.XR_HAND_JOINT_THUMB_METACARPAL,
                    XrHandJoint.XR_HAND_JOINT_THUMB_PROXIMAL,
                    XrHandJoint.XR_HAND_JOINT_THUMB_DISTAL),

                XrFinger.XR_HAND_INDEX => (
                    XrHandJoint.XR_HAND_JOINT_INDEX_PROXIMAL,
                    XrHandJoint.XR_HAND_JOINT_INDEX_INTERMEDIATE,
                    XrHandJoint.XR_HAND_JOINT_INDEX_DISTAL),

                XrFinger.XR_HAND_MIDDLE => (
                    XrHandJoint.XR_HAND_JOINT_MIDDLE_PROXIMAL,
                    XrHandJoint.XR_HAND_JOINT_MIDDLE_INTERMEDIATE,
                    XrHandJoint.XR_HAND_JOINT_MIDDLE_DISTAL),

                XrFinger.XR_HAND_RING => (
                    XrHandJoint.XR_HAND_JOINT_RING_PROXIMAL,
                    XrHandJoint.XR_HAND_JOINT_RING_INTERMEDIATE,
                    XrHandJoint.XR_HAND_JOINT_RING_DISTAL),

                XrFinger.XR_HAND_PINKY => (
                    XrHandJoint.XR_HAND_JOINT_LITTLE_PROXIMAL,
                    XrHandJoint.XR_HAND_JOINT_LITTLE_INTERMEDIATE,
                    XrHandJoint.XR_HAND_JOINT_LITTLE_DISTAL),

                _ => throw new ArgumentOutOfRangeException(nameof(id), id, null)
            };

        public override string ToString()
        {
            var sb = new StringBuilder();
            sb.Append($"WRIST : [ {rootPart} ]\n");
            sb.Append($"THUMB PROXIMAL : [ {thumbMiddle} ]\n");
            sb.Append($"THUMB DISTAL : [ {thumbTop} ]\n");
            sb.Append($"INDEX PROXIMAL : [ {indexBase} ]\n");
            sb.Append($"INDEX INTERMEDIATE : [ {indexMiddle} ]\n");
            sb.Append($"INDEX DISTAL : [ {indexTop} ]\n");
            sb.Append($"MIDDLE PROXIMAL : [ {middleBase} ]\n");
            sb.Append($"MIDDLE INTERMEDIATE : [ {middleMiddle} ]\n");
            sb.Append($"MIDDLE DISTAL : [ {middleTop} ]\n");
            sb.Append($"RING PROXIMAL : [ {ringBase} ]\n");
            sb.Append($"RING INTERMEDIATE : [ {ringMiddle} ]\n");
            sb.Append($"RING DISTAL : [ {ringTop} ]\n");
            sb.Append($"PINKY PROXIMAL : [ {pinkyBase} ]\n");
            sb.Append($"PINKY INTERMEDIATE : [ {pinkyMiddle} ]\n");
            sb.Append($"PINKY DISTAL : [ {pinkyTop} ]\n");
            return sb.ToString();
        }
    }
}