// /******************************************************************************
//  * File: HandVisualizer.cs
//  * Copyright (c) 2023 Qualcomm Technologies, Inc. and/or its subsidiaries. All rights reserved.
//  *
//  * Confidential and Proprietary - Qualcomm Technologies, Inc.
//  *
//  ******************************************************************************/

using System.Collections;
using UnityEngine;

namespace QCHT.Interactions.Hands
{
    public class HandVisualizer : MonoBehaviour, IHandSkinnable, IHideable 
    {
        private static readonly int s_globalAlpha = Shader.PropertyToID("_Alpha");
        private static readonly int s_userAlpha = Shader.PropertyToID("_OverrideAlpha");

        private const float kFadeDuration = 0.33f;

        private HandFadeOut _fadeOutInstance; // Use one single fade out instance
        
        [SerializeField] protected SkinnedMeshRenderer meshRenderer;
        
        private MaterialPropertyBlock _propertyBlock;

        private Coroutine _coroutine;

        protected void Awake() {
            _propertyBlock = new MaterialPropertyBlock();
            _fadeOutInstance = CreateFadeInstance();
        }

        protected void OnDestroy() {
            Destroy(_fadeOutInstance);
        }

        #region IHandSkinnable
        
        public void SetSkin(HandSkin skin) {
            if (meshRenderer == null)
                return;
            
            meshRenderer.sharedMesh = skin.MainMesh;
            meshRenderer.material = skin.MainMaterial;
        }
        
        #endregion

        #region IHandFadable
        
        public void Hide() {
            if (_fadeOutInstance) {
                _fadeOutInstance.gameObject.SetActive(true);
                var fadeOutInstanceTransform = _fadeOutInstance.transform;
                var thisTransform = transform;
                fadeOutInstanceTransform.position = thisTransform.position;
                fadeOutInstanceTransform.rotation = thisTransform.rotation;
                _fadeOutInstance.TakeSnapShot(meshRenderer);
                _fadeOutInstance.StartFading(GetAlpha()); 
            }
            
            SetAlpha(0f); // Set alpha to 0, for next show call
        }

        public void Show() {
            if (_fadeOutInstance) {
                _fadeOutInstance.gameObject.SetActive(false);
            }
            if (_coroutine != null) StopCoroutine(_coroutine); 
            _coroutine = StartCoroutine(FadeInAsync());
        }

        private static HandFadeOut CreateFadeInstance() {
            // Instantiate fade out instance
            var go = new GameObject("FadeOutHand") {
                hideFlags = HideFlags.HideInHierarchy
            };
            return go.AddComponent<HandFadeOut>();
        }
        
        protected virtual IEnumerator FadeInAsync() {
            yield return Fade(1f);
        }
        
        protected IEnumerator Fade(float targetAlpha)
        {
            var startAlpha = GetAlpha();
            float time = 0;
            while (time < kFadeDuration) {
                time += Time.deltaTime;
                var dt = time / kFadeDuration;
                var alpha = Mathf.Lerp(startAlpha, targetAlpha, dt);
                SetAlpha(alpha);
                yield return null; // Wait for next frame
            }
        }
        
        protected void SetAlpha(float alpha) {
            _propertyBlock.SetFloat(s_globalAlpha, alpha);
            meshRenderer.SetPropertyBlock(_propertyBlock);
        }
        
        protected float GetAlpha() {
            meshRenderer.GetPropertyBlock(_propertyBlock);
            return _propertyBlock.GetFloat(s_globalAlpha);
        }
        
        protected void SetOptionalAlpha(float alpha) {
            _propertyBlock.SetFloat(s_userAlpha, alpha);
            meshRenderer.SetPropertyBlock(_propertyBlock);
        }
        
        #endregion
    }
}