// /******************************************************************************
//  * File: QCHTHandPresenter.cs
//  * Copyright (c) 2023 Qualcomm Technologies, Inc. and/or its subsidiaries. All rights reserved.
//  *
//  * Confidential and Proprietary - Qualcomm Technologies, Inc.
//  *
//  ******************************************************************************/

using System;
using System.Collections;
using QCHT.Core;
using QCHT.Interactions.Distal;
using QCHT.Interactions.Hands.VFF;
using QCHT.Interactions.Proximal;
using Unity.XR.CoreUtils;
using UnityEngine;
using UnityEngine.Serialization;

namespace QCHT.Interactions.Hands
{
    [Obsolete("This item is now obsolete and will be removed in future release, please consider updating your project. More information: https://docs.spaces.qualcomm.com/unity/setup/SetupGuideUnity.html",false)]
    public sealed class QCHTHandPresenter : MonoBehaviour, IInteractionController
    {
        private static readonly int s_globalAlpha = Shader.PropertyToID("_Alpha");
        private static readonly int s_userAlpha = Shader.PropertyToID("_OverrideAlpha");

        [Header("General")]
        [SerializeField] private bool isLeft;
        [SerializeField] private HandDriver mainHand;
        [SerializeField] private HandDriver ghostHand;

        [SerializeField] private SkinnedMeshRenderer mainSkinnedMeshRenderer;
        [SerializeField] private SkinnedMeshRenderer ghostSkinnedMeshRenderer;

        [SerializeField, Range(0.01f, 5f)] private float handScaleMultiplier = 1f;

        [Header("Proximal interaction")]
        [Tooltip(
            "Triggering: this one is used for distal and proximal interaction, but also for the control box and snapping features. \n \nColliding (for VFF): choose this interaction type if you want natural interaction using physics. \n \nNone: this one only displays a ghost hand avatar.")]
        [SerializeField] private XrHandInteractionType interactionType;

        [Header("Distal interactions (QCHT)")]
        [SerializeField, Tooltip("Should enable physics raycast?\nIt needs a QCHTInputModule to process raycasting.")]
        private bool enablePhysicsRaycast = true;
        [SerializeField] private QCHTPhysicsRaycaster raycaster;

        [SerializeField, Tooltip("Should enable the qcht ray visual?")]
        private bool enableRaycastDisplay = true;
        [SerializeField] private QCHTRay ray;

        [Header("Display")]
        [SerializeField] private bool displayMainHand = true;
        [SerializeField] private bool displayGhostHand = true;
        [SerializeField] private bool displayMainToGhostLine = true;
        [Tooltip("Setup the fade-out duration when the hand is not detected anymore.")]
        [SerializeField] private float fadeDuration = 0.33f;
        [Tooltip(
            "Determines the distance at which the ghost hand will be displayed once the default hand avatar is being physically constrained by the object.")]
        [SerializeField, MinMax(0.01f, 1f)]
        private Vector2 distanceBlendAlpha = new Vector2(0.01f, 1f);

        [FormerlySerializedAs("defaultSkin")]
        [Header("Skin")]
        [SerializeField] private HandSkin handSkin;

        private MaterialPropertyBlock _mainMaterialPropertyBlock;
        private MaterialPropertyBlock _ghostMaterialPropertyBlock;
        private Coroutine _fadeRoutine;

        private bool _handDriveAllowed = true;
        
        public XrHandInteractionType InteractionType
        {
            set
            {
                if (value == interactionType)
                    return;

                interactionType = value;
                SetupHandType(interactionType);
            }

            get => interactionType;
        }
        
        private bool _detected = true;
        
        private bool IsDetected
        {
            set
            {
                if (_detected == value)
                    return;

                _detected = value;
                if (!_handDriveAllowed) return;
                ActivateHand(_detected);
            }
            get => _detected;
        }

        public bool EnablePhysicsRaycast
        {
            set => enablePhysicsRaycast = value;
            get => enablePhysicsRaycast;
        }

        public bool EnableRayDisplay
        {
            set => enableRaycastDisplay = value;
            get => enableRaycastDisplay;
        }
        
        public bool DisplayMainHand
        {
            set => displayMainHand = value;
            get => displayMainHand;
        }

        public bool DisplayGhostHand
        {
            set => displayGhostHand = value;
            get => displayGhostHand;
        }

        public bool DisplayMainToGhostLine
        {
            set => displayMainToGhostLine = value;
            get => displayMainToGhostLine;
        }

        public Pose MainWorldPose
        {
            set
            {
                var partTransform = mainHand.RootPart.transform;
                partTransform.position = value.position;
                partTransform.rotation = value.rotation;
            }
            get
            {
                var partTransform = mainHand.RootPart.transform;
                return new Pose(partTransform.position, partTransform.rotation);
            }
        }

        public Pose GhostWorldPose
        {
            get
            {
                var partTransform = ghostHand.RootPart.transform;
                return new Pose(partTransform.position, partTransform.rotation);
            }
        }

        public HandSkin HandSkin
        {
            set
            {
                if (handSkin == value) return;

                handSkin = value;
                
                if (!handSkin)
                {
                    mainSkinnedMeshRenderer.material = null;
                    mainSkinnedMeshRenderer.sharedMesh = null;
                    ghostSkinnedMeshRenderer.material = null;
                    ghostSkinnedMeshRenderer.sharedMesh = null;
                    return;
                }

                mainSkinnedMeshRenderer.sharedMesh = handSkin.MainMesh;
                mainSkinnedMeshRenderer.material = handSkin.MainMaterial;
                ghostSkinnedMeshRenderer.sharedMesh = !handSkin.GhostMesh ? handSkin.GhostMesh : handSkin.MainMesh;
                ghostSkinnedMeshRenderer.material = handSkin.GhostMaterial;
            }

            get => handSkin;
        }
        
        #region MonoBehaviour Functions

        private void Awake()
        {
            _mainMaterialPropertyBlock = new MaterialPropertyBlock();
            _ghostMaterialPropertyBlock = new MaterialPropertyBlock();
            mainSkinnedMeshRenderer.GetPropertyBlock(_mainMaterialPropertyBlock);
            ghostSkinnedMeshRenderer.GetPropertyBlock(_ghostMaterialPropertyBlock);
        }

        private void OnEnable()
        {
            QCHTProximalInteractor.RegisterController(this, isLeft ? HandType.Left : HandType.Right);
        }

        private void OnDisable()
        {
            // Force releasing interaction if the Controller is disabled
            ReleaseInteraction();

            QCHTProximalInteractor.UnRegisterController(isLeft ? HandType.Left : HandType.Right);
        }

        private void Start()
        {
            SetupHandType(interactionType);
        }

        private void Update()
        {
            IsDetected = QCHTInput.IsHandDetected(isLeft);

            UpdateHandDisplay();
            
            var canRaycast = IsDetected && QCHTInput.CanPhysicsRaycast(isLeft);
            UpdateRaycaster(canRaycast);
            UpdateRaycastDisplay();
            
            var dist = GetNormalizedDistanceInMinMaxRange();
            SetGhostHandAlpha(dist);
        }

        #endregion

        private void ActivateHand(bool on)
        {
            if (_fadeRoutine != null)
                StopCoroutine(_fadeRoutine);

            _fadeRoutine = StartCoroutine(Fade(on ? 1f : 0f));
        }

        public void AllowHandDisplay(bool on)
        {
            if (_handDriveAllowed == on) return;
            _handDriveAllowed = on;
            if (_detected) ActivateHand(on);
        }

        public void SetupHandTypeFromId(int typeId)
        {
            try
            {
                SetupHandType((XrHandInteractionType) typeId);
            }
            catch (ArgumentOutOfRangeException)
            {
                Debug.LogError($"[HandPresenter:SetupHandType] Can't set type id {typeId}");
            }
        }

        public void SetupHandType(XrHandInteractionType handInteractionType)
        {
            switch (handInteractionType)
            {
                case XrHandInteractionType.XR_HAND_TRIGGERING:
                    SetupTriggering();
                    break;

                case XrHandInteractionType.XR_HAND_COLLIDING:
                    SetupColliding();
                    break;

                case XrHandInteractionType.XR_HAND_NONE:
                    SetupNone();
                    break;

                default:
                    throw new ArgumentOutOfRangeException();
            }
        }

        private void SetupNone()
        {
            mainHand.gameObject.SetActive(_detected);
            ghostHand.gameObject.SetActive(false);
            Root.TriggerCollisionsDetection(false);
        }

        private void SetupTriggering()
        {
            mainHand.gameObject.SetActive(_detected);
            ghostHand.gameObject.SetActive(_detected && displayGhostHand);
            Root.TriggerCollisionsDetection(true);
        }

        private void SetupColliding()
        {
            mainHand.gameObject.SetActive(_detected);
            ghostHand.gameObject.SetActive(_detected && displayGhostHand);
        }

        private void SetMainHandAlpha(float alpha)
        {
            _mainMaterialPropertyBlock.SetFloat(s_globalAlpha, alpha);
            mainSkinnedMeshRenderer.SetPropertyBlock(_mainMaterialPropertyBlock);
        }

        private void SetGhostHandAlpha(float alpha)
        {
            _ghostMaterialPropertyBlock.SetFloat(s_globalAlpha, alpha);
            ghostSkinnedMeshRenderer.SetPropertyBlock(_ghostMaterialPropertyBlock);
        }

        public void SetMainHandOptionalAlpha(float alpha)
        {
            _mainMaterialPropertyBlock.SetFloat(s_userAlpha, alpha);
            mainSkinnedMeshRenderer.SetPropertyBlock(_mainMaterialPropertyBlock);
        }

        public void SetGhostHandOptionalAlpha(float alpha)
        {
            _ghostMaterialPropertyBlock.SetFloat(s_userAlpha, alpha);
            ghostSkinnedMeshRenderer.SetPropertyBlock(_ghostMaterialPropertyBlock);
        }

        private void UpdateHandDisplay()
        {
            mainSkinnedMeshRenderer.enabled = displayMainHand;
            ghostSkinnedMeshRenderer.enabled = displayGhostHand;
        }

        private void UpdateRaycastDisplay()
        {
            if (!ray) return;
            var eRayD = enableRaycastDisplay;
            ray.gameObject.SetActive(eRayD);
        }

        private void UpdateRaycaster(bool canRaycast)
        {
            if (!raycaster) return;
            var eRayR = enablePhysicsRaycast && canRaycast;
            raycaster.enabled = eRayR;
            raycaster.gameObject.SetActive(eRayR);
        }
        
        private float GetNormalizedDistanceInMinMaxRange()
        {
            var distance = Vector3.Distance(MainWorldPose.position, GhostWorldPose.position);
            distance = (distance - distanceBlendAlpha.x) / distanceBlendAlpha.y - distanceBlendAlpha.x;
            return Mathf.Clamp01(distance);
        }

        private IEnumerator Fade(float targetAlpha)
        {
            if (targetAlpha > 0)
            {
                switch (interactionType)
                {
                    case XrHandInteractionType.XR_HAND_TRIGGERING:
                        mainHand.gameObject.SetActive(_detected);
                        ghostHand.gameObject.SetActive(_detected && displayGhostHand);
                        break;

                    case XrHandInteractionType.XR_HAND_COLLIDING:
                        mainHand.gameObject.SetActive(_detected);
                        ghostHand.gameObject.SetActive(_detected && displayGhostHand);
                        break;

                    case XrHandInteractionType.XR_HAND_NONE:
                        mainHand.gameObject.SetActive(_detected);
                        ghostHand.gameObject.SetActive(false);
                        break;
                }

                mainHand.enabled = true;
                ghostHand.enabled = true;
            }
            else
            {
                if (_handDriveAllowed)
                {
                    mainHand.enabled = false;
                    ghostHand.enabled = false;
                }
            }

            mainSkinnedMeshRenderer.GetPropertyBlock(_mainMaterialPropertyBlock);
            ghostSkinnedMeshRenderer.GetPropertyBlock(_ghostMaterialPropertyBlock);
            var startMainAlpha = _mainMaterialPropertyBlock.GetFloat(s_globalAlpha);
            var startGhostAlpha = _ghostMaterialPropertyBlock.GetFloat(s_globalAlpha);
            float time = 0;

            while (time < fadeDuration)
            {
                time += Time.deltaTime;
                var dt = time / fadeDuration;
                var mainAlpha = Mathf.Lerp(startMainAlpha, targetAlpha, dt);
                var ghostAlpha = Mathf.Lerp(startGhostAlpha, targetAlpha, dt);
                var alpha = GetNormalizedDistanceInMinMaxRange();
                SetMainHandAlpha(mainAlpha);
                SetGhostHandAlpha(ghostAlpha * alpha);
                yield return null;
            }

            if (targetAlpha > 0)
                yield break;

            if (!_handDriveAllowed)
                yield break;

            mainHand.gameObject.SetActive(false);
            ghostHand.gameObject.SetActive(false);
        }

        #region IInteractionController

        public HandPhysicsPart Root;
        public XrHandedness Handedness => isLeft ? XrHandedness.XR_HAND_LEFT : XrHandedness.XR_HAND_RIGHT;
        public HandDriver Driver => ghostHand;
        public Pose Pose {
            get => ghostHand.transform.GetWorldPose();
            set {
                mainHand.transform.position = value.position;
                mainHand.transform.rotation = value.rotation;
            }
        }

        public GameObject TriggeredObject
        {
            get
            {
                return null;
            }
        }
        
        public void OnBeginInteraction(bool shouldStopTracking)
        {
        }

        public void OnEndInteraction()
        {
        }

        public void ReleaseInteraction()
        {
            OnEndInteraction();
        }
        
        #endregion
    }
}