// /******************************************************************************
//  * File: ProximalInteractable.cs
//  * Copyright (c) 2023 Qualcomm Technologies, Inc. and/or its subsidiaries. All rights reserved.
//  *
//  * Confidential and Proprietary - Qualcomm Technologies, Inc.
//  *
//  ******************************************************************************/

using System;
using System.Collections;
using System.Linq;
using QCHT.Interactions.Hands;
using Unity.XR.CoreUtils;
using UnityEngine;
using UnityEngine.Events;

namespace QCHT.Interactions.Proximal
{
    public enum InteractionType
    {
        Grabbable,
        Constraint,
        ConstraintMovingPart
    }

    [ExecuteInEditMode]
    [Obsolete("This item is now obsolete and will be removed in future release, please consider updating your project. More information: https://docs.spaces.qualcomm.com/unity/setup/SetupGuideUnity.html " +
              "Use a XR Grabbable component and XRFilter to filter grab and pinch actions instead.")]
    public sealed class ProximalInteractable : MonoBehaviour, IProximalEnterHandler, IProximalExitHandler,
        IBeginProximalGrabHandler, IEndProximalGrabHandler
    {
        [Header("Object")]
        [SerializeField] private InteractionType interactionType;
        [SerializeField, Tooltip("Hide controller when interacting with this object")]
        private bool hideWhenInteract;
        [SerializeField] private Transform movablePart;
        [Space]
        [SerializeField] private FreeAxis moveFreeAxis;
        [SerializeField] private Vector2 moveLimits = new Vector2(-.5f, .5f);
        [SerializeField, Range(0f, 1f)] private float moveValue;
        [Space]
        [SerializeField] private FreeAxis rotationFreeAxis;
        [SerializeField, MinMax(0f, 360f)] private Vector2 rotationLimits = new Vector2(0, 360f);
        [SerializeField, Range(0f, 1f)] private float rotationValue;
        [SerializeField] private bool fromWristRotation;
        [SerializeField] private bool inverted = true;
        [SerializeField] private bool clamped = true;
        [Space]
        [SerializeField] private bool disableCollisionsWhenInteracting = true;
        [SerializeField] private bool animateWhenSnap = true;

        [Header("Break")]
        [SerializeField] private float breakDistance;
        [SerializeField] private bool releaseOnBreak;
        [SerializeField] private UnityEvent OnBreakEvent;

        [Header("Settings")]
        [SerializeField] private bool grabbableWithTwoHands = true;
        [SerializeField] private float freeGrabbingSpeed = 15f;
        [SerializeField, MinMax(0f, 10f)] private Vector2 maxScaleMultiplier = new Vector2(.5f, 2f);

        [Header("Callbacks"), Space]
        public UnityEvent<InteractionData> OnProximalEnterEvent = new UnityEvent<InteractionData>();
        public UnityEvent<InteractionData> OnProximalExitEvent = new UnityEvent<InteractionData>();
        public UnityEvent<InteractionData> OnProximalGrabEvent = new UnityEvent<InteractionData>();
        public UnityEvent<InteractionData> OnProximalReleaseEvent = new UnityEvent<InteractionData>();

        private const float SnappingDuration = 0.1f;

        private Rigidbody _rigidbody;
        private ConfigurableJoint _joint;
        private Collider[] _colliders; // Collider that are not trigger

        // Grabbing offset
        private Pose _leftGrabPoseOffset;
        private Pose _rightGrabPoseOffset;
        
        private Coroutine _leftGrabRoutine;
        private Coroutine _rightGrabRoutine;
        
        public bool IsGrabbed { get; private set; }
        public Transform MovablePart => movablePart ? movablePart : transform;

        public float MoveValue
        {
            get => moveValue;
            set => moveValue = value;
        }

        public float RotationValue
        {
            get => rotationValue;
            set => rotationValue = value;
        }

        public Vector2 LimitsValues
        {
            get => moveLimits;
            set => moveLimits = value;
        }

        #region MonoBehaviour Functions

        public void Awake()
        {
            _rigidbody = GetComponent<Rigidbody>();
            _colliders = GetComponentsInChildren<Collider>(true).Where(col => !col.isTrigger).ToArray();
        }

        public void Update()
        {
            if (interactionType == InteractionType.ConstraintMovingPart)
            {
                var pose = UpdatePositionFromValues();
                var t = MovablePart;
                t.localPosition = pose.position;
                t.localRotation = pose.rotation;
            }
        }

        public void OnDisable() => ReleaseAllGrab();

        #endregion

        private Pose UpdatePositionFromValues()
        {
            var t = MovablePart;
            var pose = new Pose();
            
            // Move
            if (moveFreeAxis != FreeAxis.None)
            {
                var moveAxis = moveFreeAxis switch
                {
                    FreeAxis.X => Vector3.right,
                    FreeAxis.Y => Vector3.up,
                    FreeAxis.Z => Vector3.forward,
                    _ => Vector3.zero
                };

                var mask = Vector3.one - moveAxis;
                var lerp = Mathf.Lerp(moveLimits.x, moveLimits.y, moveValue);
                mask.Scale(t.localPosition);
                pose.position = lerp * moveAxis + mask;
            }

            // Rotate
            if (rotationFreeAxis != FreeAxis.None)
            {
                var rotAxis = rotationFreeAxis switch
                {
                    FreeAxis.X => inverted ? Vector3.left : Vector3.right,
                    FreeAxis.Y => inverted ? Vector3.down : Vector3.up,
                    FreeAxis.Z => inverted ? Vector3.back : Vector3.forward,
                    _ => Vector3.zero
                };

                var angle = rotationLimits.x + rotationValue * (rotationLimits.y - rotationLimits.x);
                pose.rotation = Quaternion.AngleAxis(angle, rotAxis);
            }

            return pose;
        }

        #region Proximal Events

        public void OnProximalEnter(InteractionData interactionData) => OnProximalEnterEvent?.Invoke(interactionData);

        public void OnProximalExit(InteractionData interactionData) => OnProximalExitEvent?.Invoke(interactionData);

        public void OnBeginGrab(InteractionData interactionData)
        {
            // Release snapping when the grabbable interactable is already grabbed
            if (IsGrabbed && interactionType == InteractionType.Grabbable)
                ReleaseAllGrab();

            var stopTracking = interactionType == InteractionType.Constraint ||
                               interactionType == InteractionType.ConstraintMovingPart;
            
            interactionData.Controller.OnBeginInteraction(stopTracking);
            
            GetPoseForFreeGrabbing(interactionData);

            ref var routine = ref interactionData.Controller.Handedness == XrHandedness.XR_HAND_LEFT
                ? ref _leftGrabRoutine
                : ref _rightGrabRoutine;
            if (routine != null) StopCoroutine(routine);
            routine = StartCoroutine(GrabAsync(interactionData));

            if (disableCollisionsWhenInteracting)
                ToggleObjectCollision(false);

            OnProximalGrabEvent.Invoke(interactionData);
        }

        public void OnEndGrab(InteractionData interactionData)
        {
            if (disableCollisionsWhenInteracting)
                ToggleObjectCollision(true);

            if (interactionData != null && interactionData.Controller != null)
            {
                var controllerType = interactionData.Controller.Handedness;
                ref var routine = ref controllerType == XrHandedness.XR_HAND_LEFT
                    ? ref _leftGrabRoutine
                    : ref _rightGrabRoutine;

                if (routine != null)
                {
                    StopCoroutine(routine);
                    routine = null;
                }

                if (interactionData.Controller is var controller && controller != null)
                {
                    if (animateWhenSnap && isActiveAndEnabled)
                        routine = StartCoroutine(UnGrabAsync(interactionData));

                    controller.OnEndInteraction();
                }

                ref var poseOffset = ref controllerType == XrHandedness.XR_HAND_LEFT
                    ? ref _leftGrabPoseOffset
                    : ref _rightGrabPoseOffset;
                poseOffset = Pose.identity;
            }

            OnProximalReleaseEvent.Invoke(interactionData);
        }

        #endregion

        #region Pose/Offsets Calculations

        // private void TryGetPoseForSnapping(InteractionData interactionData)
        // {
        //     var snapData = interactionData.Data?.SnapData;
        //     if (snapData == null)
        //     {
        //         GetPoseForFreeGrabbing(interactionData);
        //         return;
        //     }
        //
        //     try
        //     {
        //         var data = snapData.Value;
        //         data.HandPose.Handedness = interactionData.Controller.Handedness;
        //         ref var poseOffset = ref data.HandPose.Handedness == XrHandedness.XR_HAND_LEFT
        //             ? ref _leftGrabPoseOffset
        //             : ref _rightGrabPoseOffset;
        //         var refPose = interactionData.Controller.Pose;
        //         data.GetPosePoseOffset(data, MovablePart, refPose, ref poseOffset);
        //     }
        //     catch (Exception)
        //     {
        //         // Fall back to free pose
        //         GetPoseForFreeGrabbing(interactionData);
        //     }
        // }

        private void GetPoseForFreeGrabbing(InteractionData interactionData)
        {
            var t = MovablePart;
            var rotation = t.rotation;
            var rTransform = interactionData.Controller.Pose;

            ref var poseOffset = ref interactionData.Controller.Handedness == XrHandedness.XR_HAND_LEFT
                ? ref _leftGrabPoseOffset
                : ref _rightGrabPoseOffset;

            if (hideWhenInteract)
            {
                poseOffset = Pose.identity;
                return;
            }

            poseOffset.rotation = Quaternion.Inverse(rotation) * rTransform.rotation;
            poseOffset.position = Quaternion.Inverse(rotation) * (rTransform.position - t.position);
        }

        #endregion

        public void ReleaseGrab() => ReleaseAllGrab();
        
        public void ReleaseAllGrab()
        {
            if (!IsGrabbed)
                return;
            
            if (_leftGrabRoutine != null) StopCoroutine(_leftGrabRoutine);
            if (_rightGrabRoutine != null) StopCoroutine(_rightGrabRoutine);
        }
        
        private void ToggleObjectCollision(bool enable)
        {
            if (_rigidbody)
                _rigidbody.isKinematic = !enable;

            foreach (var col in _colliders)
                col.isTrigger = !enable;
        }

        private IEnumerator GrabAsync(InteractionData interactionData)
        {
            var t = MovablePart;
            var poseOffset = interactionData.Controller.Handedness == XrHandedness.XR_HAND_LEFT
                ? _leftGrabPoseOffset
                : _rightGrabPoseOffset;

            var controllerPose = interactionData.Controller.Pose;
            var lastControllerPose = controllerPose;
            for (var time = animateWhenSnap ? 0f : SnappingDuration;; time += Time.deltaTime)
            {
                controllerPose = interactionData.Controller.Pose;
                var dt = Mathf.Clamp01(time / SnappingDuration);

                switch (interactionType)
                {
                    case InteractionType.Grabbable:
                        var tRot = controllerPose.rotation * Quaternion.Inverse(poseOffset.rotation);
                        t.rotation = Quaternion.Slerp(t.rotation, tRot, dt);
                        var tPos = controllerPose.position - t.rotation * poseOffset.position;
                        t.position = Vector3.Lerp(t.position, tPos, dt);
                        break;

                    case InteractionType.Constraint:
                        interactionData.Controller.Pose = t.TransformPose(poseOffset);
                        break;

                    case InteractionType.ConstraintMovingPart:
                        var iT = transform;
                        var controllerPos = controllerPose.position;

                        // Move
                        var moveAxis = moveFreeAxis switch
                        {
                            FreeAxis.X => iT.right,
                            FreeAxis.Y => iT.up,
                            FreeAxis.Z => iT.forward,
                            _ => Vector3.zero
                        };

                        var tToController = controllerPos - t.position;
                        var lastTtoController = lastControllerPose.position - t.position;
                        var driverDot = Vector3.Dot(moveAxis, tToController);
                        var dot = Vector3.Dot(moveAxis, lastTtoController);
                        if (dot * driverDot >= 0)
                        {
                            var newPos = moveValue + dot / (moveLimits.y - moveLimits.x);
                            moveValue = Mathf.Clamp01(newPos);
                        }

                        // Rotate
                        var rotAxis = rotationFreeAxis switch
                        {
                            FreeAxis.X => inverted ? -iT.right : iT.right,
                            FreeAxis.Y => inverted ? -iT.up : iT.up,
                            FreeAxis.Z => inverted ? -iT.forward : iT.forward,
                            _ => Vector3.zero
                        };

                        float angle;

                        if (fromWristRotation)
                        {
                            var lastControllerRotation = lastControllerPose.rotation;
                            var controllerRotation = controllerPose.rotation;
                            var a1 = Quaternion.Angle(lastControllerRotation, Quaternion.Euler(rotAxis));
                            var a2 = Quaternion.Angle(controllerRotation, Quaternion.Euler(rotAxis));
                            angle = (a2 - a1) / 180f;
                        }
                        else
                        {
                            var pLastTtoController = Vector3.ProjectOnPlane(lastTtoController, rotAxis).normalized;
                            var pTtoController = Vector3.ProjectOnPlane(tToController, rotAxis).normalized;
                            angle = Vector3.SignedAngle(pLastTtoController, pTtoController, rotAxis) / 360f;
                        }

                        var newRot = rotationValue + angle;

                        if (!clamped)
                        {
                            if (newRot > 1f) newRot -= 1f;
                            if (newRot < 0f) newRot += 1f;
                        }

                        rotationValue = Mathf.Clamp01(newRot);
                        lastControllerPose = controllerPose;
                        var movingPartPose = UpdatePositionFromValues();
                        interactionData.Controller.Pose = poseOffset.GetTransformedBy(t.TransformPose(movingPartPose));
                        break;
                }

                yield return null;
            }
        }

        private static IEnumerator UnGrabAsync(InteractionData interactionData)
        {
            if (interactionData.Controller is var hand && hand != null)
                yield break;

            for (var time = 0f; time < SnappingDuration; time += Time.deltaTime)
            {
                var dt = Mathf.Clamp01(time / SnappingDuration);
                var handPos = hand.Driver.RootTransform.localPosition;
                var handRot = hand.Driver.RootTransform.localRotation;
                handPos = Vector3.Lerp(handPos, Vector3.zero, dt);
                handRot = Quaternion.Slerp(handRot, Quaternion.identity, dt);
                hand.Driver.RootTransform.localPosition = handPos;
                hand.Driver.RootTransform.localRotation = handRot;
                yield return null;
            }
        }
    }
}