// /******************************************************************************
//  * File: QCHTProximalInteractor.cs
//  * Copyright (c) 2023 Qualcomm Technologies, Inc. and/or its subsidiaries. All rights reserved.
//  *
//  * Confidential and Proprietary - Qualcomm Technologies, Inc.
//  *
//  ******************************************************************************/

using QCHT.Core;
using QCHT.Core.Extensions;
using QCHT.Interactions.Hands;
using UnityEngine;
using UnityEngine.EventSystems;

#if UNITY_EDITOR
using UnityEditor;
#endif

namespace QCHT.Interactions.Proximal
{
    public sealed class QCHTProximalInteractor : MonoBehaviour
    {
        public enum ProximalGestureId
        {
            Pinch = GestureId.PINCH,
            Grab = GestureId.GRAB,
            Point = GestureId.POINT,
            OpenHand = GestureId.OPEN_HAND,
        }

        [Header("Gestures for proximal interaction")]
        [SerializeField]
        [Tooltip("Default gesture to start proximal interaction, IBeginGrabHandler")]
        private ProximalGestureId mainGesture = ProximalGestureId.Grab;

        [SerializeField]
        [Tooltip("A second gesture can be used to start proximal interaction.")]
        private ProximalGestureId secondaryGesture = ProximalGestureId.Pinch;

        [SerializeField]
        [Tooltip("Default gesture to release the proximal interaction")]
        private ProximalGestureId defaultReleaseGesture = ProximalGestureId.OpenHand;

        private static InteractionData s_leftInteractionData;
        private static InteractionData s_rightInteractionData;
        private static DoubleInteractionData s_doubleInteractionData;

        private void Update()
        {
            ProcessControllers();
        }

        private void ProcessControllers()
        {
            if (s_leftController != null)
                ProcessController(s_leftController, s_leftInteractionData);

            if (s_rightController != null)
                ProcessController(s_rightController, s_rightInteractionData);

            if (s_doubleInteractionData != null)
                ProcessDoubleGrab(s_doubleInteractionData);
        }

        private void ProcessController(IInteractionController controller, InteractionData interactionData)
        {
            GetDataFromController(controller, interactionData);
            ProcessHandInput(controller, interactionData);
            ProcessGrab(interactionData);
        }

        private static void GetDataFromController(IInteractionController controller, InteractionData eventData)
        {
            if (controller == null)
                return;

            var go = controller.TriggeredObject;

            if (!go)
            {
                if (eventData.ProximalEnter)
                {
                    ExecuteEvents.ExecuteHierarchy(eventData.ProximalEnter, eventData,
                        ExecuteProximalEventsExtensions.exitHandler);
                    eventData.ProximalEnter = null;
                }
            }

            if (!eventData.ProximalEnter || eventData.ProximalEnter != go)
            {
                eventData.ProximalEnter = go;

                if (eventData.ProximalEnter && !eventData.ProximalGrab)
                {
                    ExecuteEvents.ExecuteHierarchy(eventData.ProximalEnter, eventData,
                        ExecuteProximalEventsExtensions.enterHandler);
                    var dataProvider = eventData.ProximalEnter.GetComponentInChildren<IHandPoseProvider>();
                    eventData.Data = dataProvider;
                }
            }
        }

        private void ProcessHandInput(IInteractionController controller, InteractionData eventData)
        {
            var gesture = (GestureId) mainGesture;
            var releaseGesture = (GestureId) defaultReleaseGesture;
            
            if (GetGestureDownThisFrame(eventData, gesture))
            {
                if (!eventData.ProximalGrab)
                {
                    var grabbable = ExecuteEvents.GetEventHandler<IBeginProximalGrabHandler>(eventData.ProximalEnter);

                    if (grabbable)
                    {
                        ExecuteEvents.ExecuteHierarchy(grabbable, eventData,
                            ExecuteProximalEventsExtensions.beginGrabHandler);
                        eventData.ProximalGrab = grabbable;
                    }
                }
            }
            else if (GetReleaseGestureDownThisFrame(eventData, releaseGesture))
            {
                if (eventData.ProximalGrab)
                {
                    ExecuteEvents.ExecuteHierarchy(eventData.ProximalGrab, eventData,
                        ExecuteProximalEventsExtensions.endGrabHandler);
                    eventData.ProximalGrab = null;
                    eventData.ProximalDoubleGrab = null;
                }
            }
        }

        private static void ProcessDoubleGrab(DoubleInteractionData eventData)
        {
            var leftData = eventData.LeftData;
            var rightData = eventData.RightData;

            // Totally stops double grab when one hand stopped double grabbing
            if (leftData.ProximalDoubleGrab != rightData.ProximalDoubleGrab)
            {
                leftData.ProximalDoubleGrab = null;
                rightData.ProximalDoubleGrab = null;

                // Get back to proximal for the hand which still grabbing
                var data = leftData.ProximalGrab ? leftData : rightData;
                if (data.ProximalGrab)
                {
                    ExecuteEvents.ExecuteHierarchy(data.ProximalGrab, data,
                        ExecuteProximalEventsExtensions.beginGrabHandler);
                }
            }

            // Begin double grabbing
            if (!leftData.ProximalDoubleGrab && leftData.ProximalGrab &&
                leftData.ProximalGrab == rightData.ProximalGrab)
            {
                var doubleGrab = ExecuteEvents.ExecuteHierarchy(leftData.ProximalGrab, s_doubleInteractionData,
                    ExecuteProximalEventsExtensions.beginDoubleGrabHandler);
                leftData.ProximalDoubleGrab = doubleGrab;
                rightData.ProximalDoubleGrab = doubleGrab;
            }

            // Double grabbing
            if (leftData.ProximalDoubleGrab)
            {
                var doubleGrab = ExecuteEvents.GetEventHandler<IDoubleGrabHandler>(leftData.ProximalDoubleGrab);

                if (doubleGrab)
                {
                    ExecuteEvents.Execute(doubleGrab, s_doubleInteractionData,
                        ExecuteProximalEventsExtensions.doubleGrabHandler);
                    leftData.ProximalDoubleGrab = doubleGrab;
                    rightData.ProximalDoubleGrab = doubleGrab;
                }
            }
        }

        private static void ProcessGrab(InteractionData eventData)
        {
            if (eventData.ProximalGrab && !eventData.ProximalDoubleGrab)
            {
                ExecuteEvents.Execute(eventData.ProximalGrab, eventData, ExecuteProximalEventsExtensions.grabHandler);
            }
        }

        #region static

        private static bool GetGestureDownThisFrame(InteractionData eventData, GestureId gesture)
        {
            var isLeft = eventData.Controller.Handedness == (XrHandedness) HandType.Left;
            var mainGesture = QCHTInput.GestureDown(isLeft, gesture);
            return mainGesture;
        }

        private static bool GetReleaseGestureDownThisFrame(InteractionData eventData, GestureId gesture,
            GestureId secondaryGesture = GestureId.UNKNOWN)
        {
            var isLeft = eventData.Controller.Handedness == (XrHandedness) HandType.Left;
            var hand = isLeft ? QCHTInput.LeftHand : QCHTInput.RightHand;
            var releaseThreshold = hand.GetMinDistToThumbTip();
            return releaseThreshold >= 0.055f;
        }

        private static IInteractionController s_leftController;
        private static IInteractionController s_rightController;

        public static void RegisterController(IInteractionController controller, HandType handType)
        {
            switch (handType)
            {
                case HandType.Left:
                    s_leftController = controller;
                    s_leftInteractionData = new InteractionData(EventSystem.current) {Controller = s_leftController};
                    break;

                case HandType.Right:
                    s_rightController = controller;
                    s_rightInteractionData = new InteractionData(EventSystem.current)
                        {Controller = s_rightController};
                    break;
            }

            if (s_leftController != null && s_rightController != null)
            {
                s_doubleInteractionData ??= new DoubleInteractionData(EventSystem.current);
                s_doubleInteractionData.LeftData = s_leftInteractionData;
                s_doubleInteractionData.RightData = s_rightInteractionData;
            }
        }

        public static void UnRegisterController(HandType handType)
        {
            ref var controller = ref handType == HandType.Left ? ref s_leftController : ref s_rightController;
            ref var eventData = ref handType == HandType.Left ? ref s_leftInteractionData : ref s_rightInteractionData;

            // Stop grabbing
            if (eventData.ProximalGrab)
            {
                ExecuteEvents.ExecuteHierarchy(eventData.ProximalGrab, eventData,
                    ExecuteProximalEventsExtensions.endGrabHandler);
            }

            controller = null;
            eventData = null;

            if (s_leftController == null || s_rightController == null)
                s_doubleInteractionData = null;
        }

        #endregion
    }
}