// /******************************************************************************
//  * File: XRHandInteractableSnapPoseProvider.cs
//  * Copyright (c) 2023 Qualcomm Technologies, Inc. and/or its subsidiaries. All rights reserved.
//  *
//  * Confidential and Proprietary - Qualcomm Technologies, Inc.
//  *
//  ******************************************************************************/

using System.Collections.Generic;
using System.Linq;
using QCHT.Interactions.Hands;
using UnityEngine;
using UnityEngine.XR.Interaction.Toolkit;

namespace QCHT.Interactions.Proximal
{
    [DisallowMultipleComponent]
    public class XRHandInteractableSnapPoseProvider : MonoBehaviour, IHandedness
    {
        [SerializeField] private XRGrabInteractable interactable;
        [SerializeField] private XrHandedness _handedness;
        [SerializeField] private HandMask mask;
        [SerializeField] private List<XRHandInteractableSnapPose> poses;

        public XrHandedness Handedness => _handedness;
        
        private HandData _interpolatedHandPose;
        private Matrix4x4 _interpolatedRootPose;
        
        private bool _wasDynamicAttached;
        private XRDirectInteractor _interactor;

        private void OnValidate()
        {
            interactable = interactable ? interactable : FindXRGrabInteractableComponent();
        }

        private void OnEnable() {
            interactable = interactable ? interactable : FindXRGrabInteractableComponent();
            if (interactable != null) {
                interactable.selectEntered.AddListener(OnSelectEntered);
                interactable.selectExited.AddListener(OnSelectExited);
            }
        }

        private void OnDisable() {
            if (interactable != null) {
                // Ensure unselect
                if (_interactor != null) {
                    OnSelectExited(new SelectExitEventArgs {
                        interactableObject = interactable,
                        interactorObject = _interactor
                    });
                }
                interactable.selectEntered.RemoveListener(OnSelectEntered);
                interactable.selectExited.RemoveListener(OnSelectExited);
            }
        }
        
        private XRGrabInteractable FindXRGrabInteractableComponent() {
            return gameObject.GetComponentInParent<XRGrabInteractable>(true);
        }
        
        #region XR Grab Interactable

        private void OnSelectEntered(SelectEnterEventArgs args) {
            var interactor = args.interactorObject as XRDirectInteractor;
            if (interactor == null)
                return;

            _interactor = interactor;

            var handedness = interactor.xrController.GetComponent<IHandedness>();
            if (handedness != null && handedness.Handedness != _handedness) return;
            
            var hand = interactor.xrController.model ? interactor.xrController.model.gameObject : interactor.gameObject;
            if (hand.TryGetComponent<IHandPoseOverrider>(out var driver)) {
                if (TryGetInterpolatedHandPoseFromScale(ref _interpolatedHandPose, ref _interpolatedRootPose, driver.RootScale)) {
                    driver.HandPoseOverride = _interpolatedHandPose;
                    driver.HandPoseMask = mask;
                
                    if (interactable != null) {
                        // force no dynamic attach
                        _wasDynamicAttached = interactable.useDynamicAttach;
                        interactable.useDynamicAttach = false;
                    }

                    var posOffset = _interpolatedRootPose.GetPosition();
                    posOffset.Scale(interactable.transform.lossyScale);
                    interactor.attachTransform.rotation = driver.RootPose.rotation * Quaternion.Inverse(_interpolatedRootPose.rotation);
                    interactor.attachTransform.position = driver.RootPose.position + interactor.attachTransform.rotation * -posOffset;
                }
            }
        }

        private void OnSelectExited(SelectExitEventArgs args) {
            var interactor = args.interactorObject as XRDirectInteractor;
            if (interactor == null)
                return;

            var hand = interactor.xrController.model ? interactor.xrController.model.gameObject : interactor.gameObject;
            if (hand.TryGetComponent<IHandPoseOverrider>(out var poseOverrider)) {
                poseOverrider.HandPoseOverride = null;
                poseOverrider.HandPoseMask = null;
            }
            
            if (interactable != null) {
                interactable.useDynamicAttach = _wasDynamicAttached;
            }
        }

        public bool TryGetInterpolatedHandPoseFromScale(ref HandData handData, ref Matrix4x4 rootPose, float scale)
        {
            if (poses == null || poses.Count == 0)
                return false;
            
            // Remove empty poses
            var validPoses = poses.Where(pose => pose != null).ToList();
            
            if (validPoses.Count == 0) return false;
            if (validPoses.Count == 1) {
                handData = validPoses[0].Data;
                rootPose = Matrix4x4.TRS(validPoses[0].transform.localPosition, validPoses[0].transform.localRotation, validPoses[0].Scale * Vector3.one);
                return true;
            }
            
            // Ensure poses are sorted
            validPoses.Sort((s1, s2) => s1.Scale.CompareTo(s2.Scale));
            
            var i1 = 0;
            for (var i = 0; i <= validPoses.Count - 2; i++) {
                if (scale > validPoses[i].Scale) {
                    i1 = i;
                }
            }
            
            var i2 = i1 + 1;
            var t = (scale - validPoses[i1].Scale) / (validPoses[i2].Scale - validPoses[i1].Scale);
            if (t > Mathf.Epsilon) {
                
                handData = HandData.Lerp(validPoses[i1].Data, validPoses[i2].Data, t);
                var lPos = Vector3.Lerp(validPoses[i1].transform.localPosition, validPoses[i2].transform.localPosition, t);
                var lRot = Quaternion.Lerp(validPoses[i1].transform.localRotation, validPoses[i2].transform.localRotation, t);
                var lSc = Vector3.one * scale;
                rootPose = Matrix4x4.TRS(lPos, lRot, lSc);
            }
            else {
                handData = validPoses[i1].Data;
                var lPos = validPoses[i1].transform.localPosition;
                var lRot = validPoses[i1].transform.localRotation;
                var lSc = Vector3.one * scale;
                rootPose = Matrix4x4.TRS(lPos, lRot, lSc);
            }
            
            return true;
        }
        
        #endregion
    }
}